from dataclasses import dataclass, field
from typing import Dict
import math

from streng.ppp.sections.concrete.typical_sections.slabs import Slab


@dataclass
class SlabLoadsToBeams(Slab):
    """

    .. code-block:: python
       :linenos:

       sl = SlabLoadsToBeams(l_max = 6.0,
                             l_min = 4.0,
                             slab_type = '4',
                             slab_load = 6.3)

       print(sl.beams_loads)

    """
    slab_load: float
    slab_areas: Dict = field(default_factory=dict)
    beams_loads: Dict = field(default_factory=dict)

    def __post_init__(self):
        self.calc()

    def calc(self):
        if self.slab_type == '1' or self.slab_type == '6' or self.slab_type == 1 or self.slab_type == 6:
            areas = self.get_areas_1(self.l_max, self.l_min)
        elif self.slab_type == '2a':
            areas = self.get_areas_2a(self.l_max, self.l_min)
        elif self.slab_type == '2b':
            areas = self.get_areas_2b(self.l_max, self.l_min)
        elif self.slab_type == '3a':
            areas = self.get_areas_3a(self.l_max, self.l_min)
        elif self.slab_type == '3b':
            areas = self.get_areas_3b(self.l_max, self.l_min)
        elif self.slab_type == '4' or self.slab_type == 4:
            areas = self.get_areas_4(self.l_max, self.l_min)
        elif self.slab_type == '5a':
            areas = self.get_areas_5a(self.l_max, self.l_min)
        elif self.slab_type == '5b':
            areas = self.get_areas_5b(self.l_max, self.l_min)
        else:
            areas = 0., 0., 0., 0.

        for i in range(0, 4):
            self.slab_areas[i + 1] = areas[i]
            self.beams_loads[i + 1] = areas[i] * self.slab_load

        self.beams_loads[1] /= self.l_max
        self.beams_loads[2] /= self.l_max
        self.beams_loads[3] /= self.l_min
        self.beams_loads[4] /= self.l_min

    @staticmethod
    def get_areas_1(l1, l2):
        """
        .. image paths are relative to the shpinx source folder that calls this automodule/autoclass

        Υπολογισμός εμβαδών για πλάκες τύπου 1

        .. image:: ../img/slab_loads_to_beams/01.png
            :width: 45 %
        """
        e1 = 0.5 * (l1 + (l1 - l2)) * l2 / 2
        e2 = e1
        e3 = 0.25 * l2 * l2
        e4 = e3
        return e1, e2, e3, e4

    @staticmethod
    def get_areas_2a(l1, l2):
        """
        Υπολογισμός εμβαδών για πλάκες τύπου 2a

        .. image:: ../img/slab_loads_to_beams/02a.png
            :width: 45 %
        """
        l2b = l2 / (1 + math.tan(math.pi / 6) / math.tan(math.pi / 4))
        l2a = l2 - l2b
        l1a = l2a
        # l1c = l2a
        l1b = l1 - 2 * l2a

        e1 = 0.5 * (l1 + l1b) * l2b
        e2 = 0.5 * (l1 + l1b) * l2a
        e3 = 0.5 * l2 * l1a
        e4 = e3
        return e1, e2, e3, e4

    @staticmethod
    def get_areas_2b(l1, l2):
        """
        Υπολογισμός εμβαδών για πλάκες τύπου 2b

        .. image:: ../img/slab_loads_to_beams/02b1.png
            :width: 45 %

        .. image:: ../img/slab_loads_to_beams/02b2.png
            :width: 40 %
        """
        l1a = (l2 / 2) / math.tan(math.pi / 6)
        l1c = l2 / 2
        l2a = l1c
        l2b = l1c
        l1b = l1 - l1a - l1c

        e1 = 0.5 * (l1 + l1b) * l2b
        e2 = 0.5 * (l1 + l1b) * l2a
        e3 = 0.5 * l2 * l1a
        e4 = 0.5 * l2 * l1c

        if l1b < 0:
            l1a = l1 / (1 + math.tan(math.pi / 6))
            l1b = l1 - l1a
            l2a = l1b
            l2c = l1b
            l2b = l2 - l2a - l2c

            e1 = 0.5 * l1 * l2c
            e2 = 0.5 * l1 * l2a
            e3 = 0.5 * (l2 + l2b) * l1a
            e4 = 0.5 * (l2 + l2b) * l1b

        return e1, e2, e3, e4

    @staticmethod
    def get_areas_3a(l1, l2):
        """
        Υπολογισμός εμβαδών για πλάκες τύπου 3a

        .. image:: ../img/slab_loads_to_beams/03a.png
            :width: 45 %
        """
        l2a = l2 / 2
        l2b = l2a
        l1a = l2a / math.tan(math.pi / 3)
        l1c = l1a
        l1b = l1 - l1a - l1c

        e1 = 0.5 * (l1 + l1b) * l2b
        e2 = 0.5 * (l1 + l1b) * l2a
        e3 = 0.5 * l2 * l1a
        e4 = 0.5 * l2 * l1c

        return e1, e2, e3, e4

    @staticmethod
    def get_areas_3b(l1, l2):
        """
        Υπολογισμός εμβαδών για πλάκες τύπου 3b

        .. image:: ../img/slab_loads_to_beams/03b1.png
            :width: 45 %

        .. image:: ../img/slab_loads_to_beams/03b2.png
            :width: 40 %
        """
        l2a = l2 / 2
        l2b = l2a
        l1a = l2a / math.tan(math.pi / 6)
        l1c = l1a
        l1b = l1 - l1a - l1c

        e1 = 0.5 * (l1 + l1b) * l2b
        e2 = 0.5 * (l1 + l1b) * l2a
        e3 = 0.5 * l2 * l1a
        e4 = 0.5 * l2 * l1c

        if l1b < 0:
            l1a = l1 / 2
            l1b = l1a
            l2a = l1a * math.tan(math.pi / 6)
            l2c = l2a
            l2b = l2 - l2a - l2c

            e1 = 0.5 * l1 * l2c
            e2 = 0.5 * l1 * l2a
            e3 = 0.5 * (l2 + l2b) * l1a
            e4 = 0.5 * (l2 + l2b) * l1b

        return e1, e2, e3, e4

    @staticmethod
    def get_areas_4(l1, l2):
        """
        Υπολογισμός εμβαδών για πλάκες τύπου 4

        .. image:: ../img/slab_loads_to_beams/04.png
            :width: 45 %
        """
        l2a = l2 / (1 + math.tan(math.pi / 3) / math.tan(math.pi / 4))
        l2b = l2 - l2a
        l1a = l2b
        l1c = l2a
        l1b = l1 - l1a - l1c

        e1 = 0.5 * (l1 + l1b) * l2b
        e2 = 0.5 * (l1 + l1b) * l2a
        e3 = 0.5 * l2 * l1a
        e4 = 0.5 * l2 * l1c

        return e1, e2, e3, e4

    @staticmethod
    def get_areas_5a(l1, l2):
        """
        Υπολογισμός εμβαδών για πλάκες τύπου 5a

        .. image:: ../img/slab_loads_to_beams/05a.png
            :width: 45 %
        """
        l2a = l2 / 2
        l2b = l2 - l2a
        l1a = l2b
        l1c = l2a / math.tan(math.pi / 3)
        l1b = l1 - l1a - l1c

        e1 = 0.5 * (l1 + l1b) * l2b
        e2 = 0.5 * (l1 + l1b) * l2a
        e3 = 0.5 * l2 * l1a
        e4 = 0.5 * l2 * l1c

        return e1, e2, e3, e4

    @staticmethod
    def get_areas_5b(l1, l2):
        """
        Υπολογισμός εμβαδών για πλάκες τύπου 5b

        .. image:: ../img/slab_loads_to_beams/05b1.png
            :width: 45 %

        .. image:: ../img/slab_loads_to_beams/05b2.png
            :width: 40 %
        """
        l2b = l2 / (1 + math.tan(math.pi / 3) / math.tan(math.pi / 4))
        l2a = l2 - l2b
        l1a = l2a
        l1c = l2a
        l1b = l1 - l1a - l1c

        e1 = 0.5 * (l1 + l1b) * l2b
        e2 = 0.5 * (l1 + l1b) * l2a
        e3 = 0.5 * l2 * l1a
        e4 = 0.5 * l2 * l1c

        if l1b < 0:
            l1a = l1 / 2
            l1b = l1a
            l2a = l1a
            l2c = l2a * math.tan(math.pi / 6)
            l2b = l2 - l2a - l2c

            e1 = 0.5 * l1 * l2c
            e2 = 0.5 * l1 * l2a
            e3 = 0.5 * (l2 + l2b) * l1a
            e4 = 0.5 * (l2 + l2b) * l1b

        return e1, e2, e3, e4






from dataclasses import dataclass, field
from typing import List
import numpy as np
import pandas as pd
import h5py
import matplotlib.pyplot as plt
import seaborn as sns
sns.set_style("whitegrid")


@dataclass
class FragilityCurves:
    ds_descriptions: List[str] = field(default_factory=list)
    thresholds: List[float] = field(default_factory=list)
    centrals: List[float] = field(default_factory=list)
    typology: str = ''
    description: str = ''
    shape: str = 'lognormal'
    imt: str = 'PGA'
    units: str = 'g'

    # private members for properties
    __centrals: List[float] = field(default_factory=list, init=False, repr=False)
    __n_ds: int = 0


    @classmethod
    def from_hdf5_discrete(cls, filename, group_name, dataset_name):
        with h5py.File(filename, mode='r') as f:
            group = f.get(group_name)
            dataset = group[dataset_name]

            fc = cls()
            fc.θs = list(dataset['θ'])
            fc.βs = list(dataset['β'])
            fc.thresholds = list(dataset['thresholds'])
            fc.ds_descriptions = dataset.attrs['damage state description']
            fc.typology = dataset.attrs['typology']
            fc.shape = dataset.attrs['shape']
            fc.imt = dataset.attrs['imt']
            fc.units = dataset.attrs['units']

            # fc.__calc_centrals()

            return fc




import pandas as pd
import string
import numpy as np
from dataclasses import dataclass, field
from typing import List, Dict
from streng.codes.greek.eak2000.raw.ch2.seismic_action import spectra
import openpyxl # import load_workbook
from tabulate import tabulate
from streng.phd.building_models.model2d_older.slab_loads import *
from streng.phd.building_models.model2d_older.model_input import Model2d


# D:\MyBooks\Keimeno\Κεφάλαιο3\Υλικό\FrameAnalyzer\ExcelFiles001\Frames_initial

@dataclass
class EtabsModel2d(Model2d):
    etabs_filename: str = ''
    etabs_file_text: List[str] = field(default_factory=list)


    def etabs_write_file(self):
        self._etabs_write_info()
        self._etabs_write_stories()
        self._etabs_write_diaphragm_names()
        self._etabs_write_grids()
        self._etabs_write_point_coordinates()
        self._etabs_write_line_connectivities()
        self._etabs_write_point_assigns()
        self._etabs_write_material_properties()
        self._etabs_write_frame_sections()
        self._etabs_write_line_assigns()
        self._etabs_write_static_load_cases()
        self._etabs_write_element_loads()
        self._etabs_write_point_loads()
        self._etabs_write_spectrum_function()
        self._etabs_write_spectrum_cases()
        self._etabs_write_analysis_options()

        self.etabs_file_text.append('  END')
        self.etabs_file_text.append('$ END OF MODEL FILE')

        file = open(self.etabs_filename, 'w')
        file.write('\n'.join(self.etabs_file_text))
        file.close()

    def _etabs_write_info(self):
        self.etabs_file_text.append('$ PROGRAM INFORMATION')
        self.etabs_file_text.append('  PROGRAM  "ETABS"  VERSION "9.7.2"')
        self.etabs_file_text.append('')
        self.etabs_file_text.append('$ CONTROLS')
        self.etabs_file_text.append('  UNITS  "KN"  "M"  ')
        self.etabs_file_text.append('  TITLE1  "university"  ')
        self.etabs_file_text.append('  PREFERENCE  MERGETOL 0.001')
        self.etabs_file_text.append('  RLLF  METHOD "TRIBAREAUBC97"  USEDEFAULTMIN "YES"  ')
        self.etabs_file_text.append('')


    def _etabs_write_analysis_options(self):
        self.etabs_file_text.append('$ ANALYSIS OPTIONS')
        self.etabs_file_text.append('  ACTIVEDOF "UX UZ RY"  ')
        self.etabs_file_text.append(f'  DYNAMICS  MODES {3*self.stories_total}  MODETYPE "EIGEN"  TOL 1E-09')
        self.etabs_file_text.append('  MASSOPTIONS  GRAVITY 9.81  SOURCE "LOADS"  LATERALONLY "YES"    STORYLEVELONLY "YES"  ')
        self.etabs_file_text.append('  MASSOPTIONS  LOAD "G"  FACTOR 1')
        self.etabs_file_text.append('  MASSOPTIONS  LOAD "Q"  FACTOR 0.3')
        self.etabs_file_text.append('')


    def _etabs_write_stories(self):
        self.etabs_file_text.append('$ STORIES - IN SEQUENCE FROM TOP')
        self.etabs_file_text.append(f'  STORY "STORY{self.stories_total}"  HEIGHT {self.stories_heights[self.stories_total-1]}  MASTERSTORY "Yes"')
        for i in range(self.stories_total-1, 0, -1):
            self.etabs_file_text.append(f'  STORY "STORY{i}"  HEIGHT {self.stories_heights[i-1]}  SIMILARTO "STORY{self.stories_total}"')
        self.etabs_file_text.append('  STORY "BASE"  ELEV 0')
        self.etabs_file_text.append('')

    def _etabs_write_diaphragm_names(self):
        self.etabs_file_text.append('$ DIAPHRAGM NAMES')
        for i in range(1, self.stories_total+1):
            self.etabs_file_text.append(f'  DIAPHRAGM "DIAPH{i}"    TYPE RIGID')
        self.etabs_file_text.append('')

    def _etabs_write_grids(self):
        alphabet = list(string.ascii_uppercase)
        self.etabs_file_text.append('$ GRIDS')
        self.etabs_file_text.append('  COORDSYSTEM "GLOBAL"  TYPE "CARTESIAN"  BUBBLESIZE 1.25')
        for i, val in enumerate(self.x_levels):
            self.etabs_file_text.append(f'  GRID "GLOBAL"  LABEL "{alphabet[i]}"  DIR "X"  COORD {val:.2f}  GRIDTYPE  "PRIMARY"    BUBBLELOC "DEFAULT"  GRIDHIDE "NO"')
        self.etabs_file_text.append('  GRID "GLOBAL"  LABEL "1"  DIR "Y"  COORD 0  GRIDTYPE  "PRIMARY"    BUBBLELOC "SWITCHED"  GRIDHIDE "NO" ')
        self.etabs_file_text.append('')

    def _etabs_write_point_coordinates(self):
        self.etabs_file_text.append('$ POINT COORDINATES')
        for i, val in enumerate(self.x_levels):
            self.etabs_file_text.append(f'  POINT "{i+1}"  {val:.2f} 0 ')
        self.etabs_file_text.append('')

    def _etabs_write_line_connectivities(self):
        self.etabs_file_text.append('$ LINE CONNECTIVITIES')
        for i in range(1, len(self.x_levels) +1):
            self.etabs_file_text.append(f'  LINE  "C{i}"  COLUMN  "{i}"  "{i}"  1')
        for i in range(1, len(self.x_levels)):
            self.etabs_file_text.append(f'  LINE  "B{i}"  BEAM  "{i}"  "{i+1}"  0')
        self.etabs_file_text.append('')

    def _etabs_write_point_assigns(self):
        self.etabs_file_text.append('$ POINT ASSIGNS')
        for i in range(1, len(self.x_levels) +1):
            self.etabs_file_text.append(f'  POINTASSIGN  "{i}"  "BASE"  RESTRAINT "UX UY UZ RX RY RZ" ')
            for j in range(1, self.stories_total + 1):
                self.etabs_file_text.append(f'  POINTASSIGN  "{i}"  "STORY{j}"  DIAPH "DIAPH{j}"  ')
        self.etabs_file_text.append('')

    def _etabs_write_material_properties(self):
        self.etabs_file_text.append('$ MATERIAL PROPERTIES')
        ds = self.input_data['materials_concrete']
        for i in range(len(ds)):
            _concID = ds.at[i, 'ConcID']
            _fc = ds.at[i, 'fc']
            _E= ds.at[i, 'E']
            _U= ds.at[i, 'U']
            _A= ds.at[i, 'A']
            _fy= ds.at[i, 'fy']
            _fys= ds.at[i, 'fys']
            self.etabs_file_text.append(f'  MATERIAL  "{_concID}"  M 0  W 0  TYPE "ISOTROPIC"  E {_E}  U {_U}  A {_A}')
            self.etabs_file_text.append(f'  MATERIAL  "{_concID}"  DESIGNTYPE "CONCRETE"  FY {_fy}  FC {_fc}  FYS {_fys}')
        self.etabs_file_text.append('')

    def _etabs_write_frame_sections(self):
        self.etabs_file_text.append('$ FRAME SECTIONS')
        ds = self.input_data['sections']
        for index, row in ds.iterrows():
            _section_name = index
            _element_type = row['ElementType']
            _b = row['b']
            _h = row['h']
            _hf = self.input_info['hf']
            _frame_loc = row['Frame_Loc']
            if _frame_loc == 'IN':
                _beff = _b + 8 * _hf
            else:
                _beff = _b + 3 * _hf
            _frame = row['Frame']
            if _element_type == 'BEAM':
                self.etabs_file_text.append(f'  FRAMESECTION  "{_section_name}"  MATERIAL "C20"  SHAPE "Tee"  D {_h}  B {_beff}  TF {_hf}  TW {_b}')
                self.etabs_file_text.append(f'  FRAMESECTION  "{_section_name}"  JMOD 0.1  I2MOD 0.5  I3MOD 0.5  MMOD 0  WMOD 0')
            elif _element_type == 'COLUMN':
                self.etabs_file_text.append(f'  FRAMESECTION  "{_section_name}"  MATERIAL "C20"  SHAPE "Rectangular"  D {_h}  B {_b}')
                self.etabs_file_text.append(f'  FRAMESECTION  "{_section_name}"  JMOD 0.1  MMOD 0  WMOD 0')
        self.etabs_file_text.append('')

    def _etabs_write_line_assigns(self):
        self.etabs_file_text.append('$ LINE ASSIGNS')
        _dict = self.input_data['elements'].to_dict('index')
        for row in _dict:
            _line = _dict[row]['Line_Etabs']
            _element_type = _dict[row]['ElementType']
            _storey = _dict[row]['Storey']
            _geometry_name = _dict[row]['GeometryName']
            if _element_type == 'BEAM':
                self.etabs_file_text.append(f'  LINEASSIGN  "{_line}"  "STORY{_storey}"  SECTION "{_geometry_name}"  ANG  0  MAXSTASPC 0.5  LENGTHOFFI 0  LENGTHOFFJ 0  RIGIDZONE 1  CARDINALPT 8    MESH "POINTSANDLINES"')
            elif 'COLUMN' in _element_type:
                self.etabs_file_text.append(f'  LINEASSIGN  "{_line}"  "STORY{_storey}"  SECTION "{_geometry_name}"  ANG  0  MINNUMSTA 3  LENGTHOFFI 0  LENGTHOFFJ 0  RIGIDZONE 1  MESH "POINTSANDLINES"')
        self.etabs_file_text.append('')

    def _etabs_write_static_load_cases(self):
        self.etabs_file_text.append('$ STATIC LOADS')
        self.etabs_file_text.append('  LOADCASE "G"  TYPE  "DEAD"  SELFWEIGHT  0')
        self.etabs_file_text.append('  LOADCASE "Q"  TYPE  "LIVE"  SELFWEIGHT  0')
        self.etabs_file_text.append('  LOADCASE "EXSTAT"  TYPE  "QUAKE"  SELFWEIGHT  0')
        self.etabs_file_text.append('  LOADCASE "EYSTAT"  TYPE  "QUAKE"  SELFWEIGHT  0')
        self.etabs_file_text.append('')


    def _etabs_write_element_loads(self):
        self.etabs_file_text.append('$ LINE OBJECT LOADS')
        ds = self.input_data['beam_loads']
        for index, row in ds.iterrows():
            _line_etabs = row['Line_Etabs']
            _storey = row['Storey']
            _g = row['gΔ']
            _q = row['qΔ']
            self.etabs_file_text.append(f'  LINELOAD  "{_line_etabs}"  "STORY{_storey:.0f}"  TYPE "UNIFF"  DIR "GRAV"  LC "G"  FVAL {_g:.2f}')
            self.etabs_file_text.append(f'  LINELOAD  "{_line_etabs}"  "STORY{_storey:.0f}"  TYPE "UNIFF"  DIR "GRAV"  LC "Q"  FVAL {_q:.2f}')
        self.etabs_file_text.append('')

    def _etabs_write_point_loads(self):
        self.etabs_file_text.append('$ POINT OBJECT LOADS')
        ds = self.input_data['node_loads'] # .query('Storey>=1' )
        for index, row in ds.iterrows():
            _storey = row['Storey']
            _point_etabs = row['Point_Etabs']
            _G = row['G']
            _Q = row['Q']
            self.etabs_file_text.append(f'  POINTLOAD  "{_point_etabs:.0f}"  "STORY{_storey:.0f}"  TYPE "FORCE"  LC "G"    FZ -{_G:.2f}')
            self.etabs_file_text.append(f'  POINTLOAD  "{_point_etabs:.0f}"  "STORY{_storey:.0f}"  TYPE "FORCE"  LC "Q"    FZ -{_Q:.2f}')
        self.etabs_file_text.append('')

    def _etabs_write_spectrum_function(self):
        self.etabs_file_text.append('  FUNCTION "EAKSPEC"  FUNCTYPE "SPECTRUM"  DAMPRATIO 0.05  SPECTYPE "USER"  ')
        _eak_ground = self.input_info['EAK_ground']
        _eak_α = self.input_info['EAK_α']
        x = np.linspace(1e-10, 2, 201)
        y = spectra.Φd(T=x,
                       α=_eak_α*9.81,
                       γI=1.0,
                       T1=spectra.T1(_eak_ground),
                       T2=spectra.T2(_eak_ground),
                       q=3.5,
                       η=1.0,
                       θ=1.0,
                       β0=2.5)
        for i in range(0, len(x)):
            self.etabs_file_text.append(f'  FUNCTION "EAKSPEC"  TIMEVAL "{x[i]:.3f}  {y[i]:.3f}"')
        self.etabs_file_text.append('')

    def _etabs_write_spectrum_cases(self):
        self.etabs_file_text.append('  RSCASE "EX"  DAMP 0.05  DIRCOMBO "SRSS"  ANG 0')
        self.etabs_file_text.append('  RSCASE "EX"  FUNC1 "EAKSPEC"  SF1 1')
        self.etabs_file_text.append('  RSCASE "EX"  MODECOMBO "CQC"')

        self.etabs_file_text.append('')







import math
import numpy as np
from scipy import integrate  # optimize, stats
from dataclasses import dataclass

from ..common.math.numerical import xy_with_endpoints
from ..common.io.output_older import Output, NamedLogTable


@dataclass
class Spectr:
    dt: float = None
    accel: np.array = None

    def __post_init__(self):
        self.step: int = None
        self.time_total: float = None
        self.time: np.array = None
        self.velocity: np.array = None
        self.displacement: np.array = None

        self.props = {'PGA': None,
                       'PGV': None,
                       'PGD': None,
                       'timePGA': None,
                       'timePGV': None,
                       'timePGD': None,
                       'PGVdivPGA': None,
                       'aRMS': None,
                       'vRMS': None,
                       'dRMS': None,
                       'Ia': None,
                       'Ic': None,
                       'SED': None,
                       'CAV': None,
                       'Husid': None,
                       'Husid_norm': None,
                       'T_Arias5_95': None}


    @classmethod
    def calculate_time_series(cls, num_steps, num_per, acc, const, omega2, dt):
        """
        Calculates the acceleration, __velocities and displacement time series for
        the SDOF oscillator
        :param dict const:
            Constants of the algorithm
        :param np.ndarray omega2:
            Square of the oscillator period
        :returns:
            x_a = Acceleration time series
            x_v = Velocity time series
            x_d = Displacement time series
        """
        x_d = np.zeros([num_steps - 1, num_per], dtype=float)
        x_v = np.zeros_like(x_d)
        x_a = np.zeros_like(x_d)

        for k in range(0, num_steps - 1):
            dug = acc[k + 1] - acc[k]
            z_1 = const['f2'] * dug
            z_2 = const['f2'] * acc[k]
            z_3 = const['f1'] * dug
            z_4 = z_1 / dt
            if k == 0:
                b_val = z_2 - z_3
                a_val = (const['f5'] * b_val) + (const['f4'] * z_4)
            else:
                b_val = x_d[k - 1, :] + z_2 - z_3
                a_val = (const['f4'] * x_v[k - 1, :]) + \
                        (const['f5'] * b_val) + (const['f4'] * z_4)

            x_d[k, :] = (a_val * const['g1']) + (b_val * const['g2']) + \
                        z_3 - z_2 - z_1
            x_v[k, :] = (a_val * const['h1']) - (b_val * const['h2']) - z_4
            x_a[k, :] = (-const['f6'] * x_v[k, :]) - (omega2 * x_d[k, :])

        return x_a, x_v, x_d

    @classmethod
    def NigamJennings(cls, acc, periods, damping, dt):
        add_PGA0 = False
        if periods[0] == 0:
            periods = np.delete(periods, 0)
            add_PGA0 = True

        num_steps = len(acc)
        num_per = len(periods)
        omega = (2. * np.pi) / np.array(periods)
        omega2 = omega ** 2.
        omega3 = omega ** 3.
        omega_d = omega * math.sqrt(1.0 - (damping ** 2.))
        const = {'f1': (2.0 * damping) / (omega3 * dt),
                 'f2': 1.0 / omega2,
                 'f3': damping * omega,
                 'f4': 1.0 / omega_d}
        const['f5'] = const['f3'] * const['f4']
        const['f6'] = 2.0 * const['f3']
        const['e'] = np.exp(-const['f3'] * dt)
        const['s'] = np.sin(omega_d * dt)
        const['c'] = np.cos(omega_d * dt)
        const['g1'] = const['e'] * const['s']
        const['g2'] = const['e'] * const['c']
        const['h1'] = (omega_d * const['g2']) - (const['f3'] * const['g1'])
        const['h2'] = (omega_d * const['g1']) + (const['f3'] * const['g2'])
        x_a, x_v, x_d = cls.calculate_time_series(num_steps, num_per, acc, const, omega2, dt)

        spectrum = {'Sa': None,
                    'Sv': None,
                    'Sd': None,
                    'T': None,
                    'PSv': None,
                    'PSa': None,
                    'ASI': None,
                    'VSI': None,
                    'HI': None,
                    'Tp': None}
        spectrum['Sa'] = np.max(np.fabs(x_a), axis=0)
        spectrum['Sv'] = np.max(np.fabs(x_v), axis=0)
        spectrum['Sd'] = np.max(np.fabs(x_d), axis=0)
        spectrum['PSv'] = spectrum['Sd'] * omega
        spectrum['PSa'] = spectrum['Sd'] * omega2
        spectrum['T'] = periods

        if add_PGA0:
            pga = np.max(np.fabs(acc))
            periods = np.append(0, periods)
            spectrum['Sa'] = np.append(pga, spectrum['Sa'])
            spectrum['Sv'] = np.append(0, spectrum['Sv'])
            spectrum['Sd'] = np.append(0, spectrum['Sd'])
            spectrum['PSv'] = np.append(0, spectrum['PSv'])
            spectrum['PSa'] = np.append(pga, spectrum['PSa'])
            spectrum['T'] = periods

        xy_ASI = xy_with_endpoints(spectrum['T'], spectrum['Sa'], 0.1, 0.5)
        spectrum['ASI'] = integrate.trapz(xy_ASI[1], xy_ASI[0])

        xy_VSI = xy_with_endpoints(spectrum['T'], spectrum['Sv'], 0.1, 2.5)
        spectrum['VSI'] = integrate.trapz(xy_VSI[1], xy_VSI[0])

        xy_HI = xy_with_endpoints(spectrum['T'], spectrum['PSv'], 0.1, 2.5)
        spectrum['HI'] = integrate.trapz(xy_HI[1], xy_HI[0])

        spectrum['Tp'] = (float(np.where(spectrum['Sa'] == max(spectrum['Sa']))[0]) + 1) * (periods[1] - periods[0])

        return spectrum

    @staticmethod
    def show_spectra_parameters(spectrum):
        paramlog = []
        paramlog.append('Acceleration Spectrum Intensity: ASI = {:.4f}'.format(spectrum['ASI']))
        paramlog.append('Velocity Spectrum Intensity: VSI = {:.4f}'.format(spectrum['VSI']))
        paramlog.append('Housner Intensity: HI = {:.4f}'.format(spectrum['HI']))
        paramlog.append('Predominant Period: Tp = {:.4f}'.format(spectrum['Tp']))

        return '\n'.join(paramlog)



    def load_single_line_textfile(self, filename):
        self.accel = np.loadtxt(filename)

    def calc(self):
        self.steps = len(self.accel)
        self.time_total = (self.steps - 1) * self.dt

        self.time = np.arange(0., self.steps * self.dt, self.dt)

        self.calc_vel_disp()

        self.props['PGA'] = max(np.absolute(self.accel))
        self.props['PGV'] = max(np.absolute(self.velocity))
        self.props['PGD'] = max(np.absolute(self.displacement))
        self.props['timePGA'] = float(np.where(np.absolute(self.accel) == self.props['PGA'])[0]) * self.dt
        self.props['timePGV'] = float(np.where(np.absolute(self.velocity) == self.props['PGV'])[0]) * self.dt
        self.props['timePGD'] = float(np.where(np.absolute(self.displacement) == self.props['PGD'])[0]) * self.dt
        self.props['PGVdivPGA'] = self.props['PGV'] / self.props['PGA']
        self.props['aRMS'] = math.sqrt((1./self.time_total)*integrate.trapz(self.accel * self.accel, self.time ))
        self.props['vRMS'] = math.sqrt((1./self.time_total)*integrate.trapz(self.velocity * self.velocity, self.time ))
        self.props['dRMS'] = math.sqrt((1./self.time_total)*integrate.trapz(self.displacement * self.displacement, self.time ))
        self.props['Ia'] = (math.pi/(2*9.81))*integrate.trapz(self.accel * self.accel, self.time )
        self.props['Ic'] = self.props['aRMS']**1.5 * math.sqrt(self.time_total)
        self.props['SED'] = integrate.trapz(self.velocity * self.velocity, self.time)
        self.props['CAV'] = integrate.trapz(np.absolute(self.accel), self.time)
        self.props['Husid'] = math.pi / (2*9.81) * integrate.cumtrapz(self.accel * self.accel, self.time, initial=0.)
        self.props['Husid_norm'] = self.props['Husid'] / self.props['Ia']
        self.props['T_Arias5_95'] = self.get_significant_duration(0.05, 0.95)

    def get_significant_duration(self, start_percentage, end_percentage):
        t_start = float(np.interp(start_percentage, self.props['Husid_norm'], self.time))
        t_end = float(np.interp(end_percentage, self.props['Husid_norm'], self.time))
        return t_end-t_start


    def show_parameters(self):
        paramlog = []
        paramlog.append('PGA = {:.4f}'.format(self.props['PGA']))
        paramlog.append('time for PGA = {:.4f}'.format(self.props['timePGA']))
        paramlog.append('PGV = {:.4f}'.format(self.props['PGV']))
        paramlog.append('time for PGV = {:.4f}'.format(self.props['timePGV']))
        paramlog.append('PGD = {:.4f}'.format(self.props['PGD']))
        paramlog.append('time for PGD = {:.4f}'.format(self.props['timePGD']))
        paramlog.append('PGV/PGA = {:.4f}'.format(self.props['PGVdivPGA']))
        paramlog.append('aRMS = {:.4f}'.format(self.props['aRMS']))
        paramlog.append('vRMS = {:.4f}'.format(self.props['vRMS']))
        paramlog.append('dRMS = {:.4f}'.format(self.props['dRMS']))
        paramlog.append('Arias Intensity Ia = {:.4f}'.format(self.props['Ia']))
        paramlog.append('Characteristic Intensity Ic = {:.4f}'.format(self.props['Ic']))
        paramlog.append('Specific Energy Density SED = {:.4f}'.format(self.props['SED']))
        paramlog.append('Cumulative Absolute Velocity CAV = {:.4f}'.format(self.props['CAV']))
        paramlog.append('Significant duration (5-95% in Husid plot) = {:.4f}'.format(self.props['T_Arias5_95']))

        return '\n'.join(paramlog)

    def calc_vel_disp(self):
        self.velocity = self.dt * integrate.cumtrapz(self.accel, initial=0.)
        self.displacement = self.dt * integrate.cumtrapz(self.velocity, initial=0.)

    def get_spectra(self, damping, startT=0., endT=4., noTs=401):
        T = np.linspace(startT, endT, noTs)
        # self.spectrum = NigamJennings(self.accelerations, T, damping, self.dt)
        return self.NigamJennings(self.accel, T, damping, self.dt)






# -*- coding: utf-8 -*-
import os
import csv
import math
import numpy as np
import matplotlib.pyplot as plt
from scipy import integrate, optimize, stats
from shapely.geometry import LineString, MultiPoint, Point
from scipy.interpolate import interp1d
from scipy.stats import norm

def read_capacity_curves(input_file):
    # This function reads one, or a set of capacity curves

    with open(input_file, 'rU') as f:
        data = csv.reader(f)
        capacity_curves = read_SdSa_capacity_curves(data)

    return capacity_curves

def read_SdSa_capacity_curves(data):

    # This function reads Sd-Sa type of capacity curves
    periods = []
    heights = []
    gammas = []
    Sdy = []
    Say = []
    Sd = []
    Sa = []

    for line in data:
        if line[0] == 'Periods [s]':
            for value in line[1:]:
                if isNumber(value):
                    periods.append(float(value))
        if line[0] == 'Heights [m]':
            for value in line[1:]:
                if isNumber(value):
                    heights.append(float(value))
        if line[0] == 'Gamma participation factors':
            for value in line[1:]:
                if isNumber(value):
                    gammas.append(float(value))
        if line[0] == 'Sdy [m]':
            for value in line[1:]:
                if isNumber(value):
                    Sdy.append(float(value))
        if line[0] == 'Say [g]':
            for value in line[1:]:
                if isNumber(value):
                    Say.append(float(value))
        if line[0][0:2] == 'Sd' and isNumber(line[0][2]):
            subSd = []
            for value in line[1:]:
                if isNumber(value):
                    subSd.append(float(value))
            Sd.append(subSd)
        if line[0][0:2] == 'Sa' and isNumber(line[0][2]):
            subSa = []
            for value in line[1:]:
                if isNumber(value):
                    subSa.append(float(value))
            Sa.append(subSa)

    # Store all the data in the dictionary
    capacity_curves = {'type': None, 'periods': None, 'heights': None,
        'gamma': None, 'Sdy': None, 'Say': None, 'Sd': None, 'Sa': None}

    capacity_curves['type'] = 'Sd-Sa'
    capacity_curves['periods'] = periods
    capacity_curves['heights'] = heights
    capacity_curves['gamma'] = gammas
    capacity_curves['Sdy'] = Sdy
    capacity_curves['Say'] = Say
    capacity_curves['Sd'] = Sd
    capacity_curves['Sa'] = Sa

    return capacity_curves

def save_capacity_curves(capacity_curves, filename):

    if capacity_curves['type'] == 'Sd-Sa':
        save_SdSa_capacity_curves(capacity_curves, filename)

def save_SdSa_capacity_curves(capacity_curves, filename):

    no_capacity_curves = len(capacity_curves['Sd'])
    output = open(filename, 'w')
    output.write('Vb-droof,FALSE\n')
    output.write('Vb-dfloor,FALSE\n')
    output.write('Sd-Sa,TRUE\n')
    periods = 'Periods [s]'
    heights = 'Heights [m]'
    gammas = 'Gamma participation factors'
    Sdy = 'Sdy [m]'
    Say = 'Say [g]'
    for icc in range(no_capacity_curves):
        if capacity_curves['periods'] !=  None:
            periods = periods + ',' + str(capacity_curves['periods'][icc])
        if capacity_curves['heights'] !=  None:
            heights = heights + ',' + str(capacity_curves['heights'][icc])
        if capacity_curves['gamma'] !=  None:
            gammas = gammas + ',' + str(capacity_curves['gamma'][icc])
        if capacity_curves['Sdy'] !=  None:
            Sdy = Sdy + ',' + str(capacity_curves['Sdy'][icc])
        if capacity_curves['Say'] !=  None:
            Say = Say + ',' + str(capacity_curves['Say'][icc])

    output.write(periods + '\n')
    output.write(heights + '\n')
    output.write(gammas + '\n')
    output.write(Sdy + '\n')
    output.write(Say + '\n')

    for icc in range(no_capacity_curves):
        Sd = 'Sd'+str(icc+1)+' [m]'
        Sa = 'Sa'+str(icc+1)+' [g]'
        for ivalue in range(len(capacity_curves['Sd'][icc])):
            Sd = Sd + ',' + str(capacity_curves['Sd'][icc][ivalue])
            Sa = Sa + ',' + str(capacity_curves['Sa'][icc][ivalue])
        output.write(Sd + '\n')
        output.write(Sa + '\n')

    output.close()

def plot_capacity_curves(capacity_curves):

    #This function plots the capacity curves
    if capacity_curves['type'] == 'Sd-Sa':
        no_capacity_curves = len(capacity_curves['Sa'])
        for icc in range(no_capacity_curves):
            Sa = capacity_curves['Sa'][icc]
            Sd = capacity_curves['Sd'][icc]
            plt.plot(Sd, Sa, color = 'g', linewidth = 2)
        plt.plot(Sd, Sa, color = 'g', linewidth = 2,
                 label = 'individual capacity curve')
        plt.xlabel('Spectral displacement [m]', fontsize = 10)
        plt.ylabel('Spectral acceleration [g]', fontsize = 10)

    elif capacity_curves['type'] == 'Vb-dfloor' or capacity_curves['type'] == 'Vb-droof':
        no_capacity_curves = len(capacity_curves['Vb'])
        for icc in range(no_capacity_curves):
            Vb = capacity_curves['Vb'][icc]
            droof = capacity_curves['droof'][icc]
            plt.plot(droof, Vb, color = 'b', linewidth = 2)
        plt.plot(droof, Vb, color = 'b', linewidth = 2,
                 label = 'individual capacity curve')
        plt.xlabel('Roof displacement [m]', fontsize = 10)
        plt.ylabel('Base shear [kN]', fontsize = 10)

    plt.suptitle('Capacity curves')
    plt.legend(loc = 'lower right', frameon = False)
    plt.show()


def read_gmrs(folder):

    #This function reads a set of ground motion records
    #and stores them in a dictionary
    time = []
    acc = []
    dt = []
    no_points = []
    name = []
    for f in os.listdir(folder):
        if f.endswith(".csv"):
            itime, iacc = read_gmr(folder, f)
            time.append(itime)
            acc.append(iacc)
            dt.append(itime[1] - itime[0])
            no_points.append(len(iacc))
            name.append(f)

    gmrs = {'time': None, 'acc': None, 'dt': None,
        'no_points': None, 'name': None}
    gmrs['time'] = time
    gmrs['acc'] = acc
    gmrs['dt'] = dt
    gmrs['no_points'] = no_points
    gmrs['name'] = name

    return gmrs

def read_gmr(folder, gmr):

    time, acc = [], []
    with open(folder + '/' + gmr) as f:
        for line in f.readlines():
            line = line.split(',')
            time.append(float(line[0]))
            acc.append(float(line[1])*9.81)

    return time, acc
    
def evaluate_optimal_IM(gmrs,PDM,minT,maxT,stepT,damage_model,damping_ratio,method):
    
    T = np.arange(minT, maxT,stepT)
    setRsquare = []
    for it in range(len(T)):
        print(str((it+1)*100/len(T))+'%')
        fragility_model = calculate_mean_fragility(gmrs, PDM, T[it], damping_ratio, 'Sa', damage_model, method)
        setRsquare.append(fragility_model['Rsquare'])

    meanRsquare = np.mean(np.array(setRsquare),axis=1)        
    bestT = T[np.argmax(meanRsquare)]
    print('The best damage-intensity correlation was achieved for Sa at T='+str(bestT)+'s')
    plot_correlation(T,setRsquare,meanRsquare,damage_model,bestT)
    
def plot_correlation(T,setRsquare,meanRsquare,damage_model,bestT):
    
    setRsquare = np.array(setRsquare)
    color_scheme = ['g', 'b', 'y', 'orangered', 'r', 'k', 'm', 'c', '0.5', '0.75']
    for iDS in range(len(damage_model['damage_states'])):
        plt.plot(T,setRsquare[:,iDS],color=color_scheme[iDS],linewidth=2,label=damage_model['damage_states'][iDS])
    
    plt.plot(T,meanRsquare,color='k',linewidth=2,linestyle='dashed',label='Mean correlation')
    plt.xlabel('T (s)', fontsize = 12)
    plt.ylabel('Coefficient of correlation - R2', fontsize = 12)
    plt.legend(loc=4)
    plt.legend(frameon=False)
    plt.show()
    
def calculate_correlation(logmeans,logstddev,cumPDM,imls):
  
    Rsquare = []
    for iDS in range(len(logmeans)):
        sigma = logstddev[iDS]
        mu = logmeans[iDS]
        expected = stats.lognorm.cdf(imls, sigma,scale=math.exp(mu)),
        observed = cumPDM[iDS+1,:]
        R = np.corrcoef(observed,expected)
        Rsquare.append(R[0][1]**2)                        
    return Rsquare

def plot_response_spectra(gmrs, minT, maxT):

    no_gmrs = len(gmrs['time'])
    damping = 0.05
    T = np.linspace(minT, maxT, 50)
    plt.figure(figsize = (15, 5))
    for igmr in range(no_gmrs):
        acc = gmrs['acc'][igmr]
        time = gmrs['time'][igmr]
        spectrum = NigamJennings(time, acc, T, damping)
    #    spectrum = NewmarkBeta(time, acc, T, damping)

        plt.subplot(1, 3, 1)
        plt.plot(T, spectrum['Sa'], color = 'b', linewidth = 2)
        plt.subplot(1, 3, 2)
        plt.plot(T, spectrum['Sd'], color = 'g', linewidth = 2)
        plt.subplot(1, 3, 3)
        plt.plot(spectrum['Sd'], spectrum['Sa'], color = 'r', linewidth = 2)

    plt.subplot(1, 3, 1)
    plt.xlabel('Periods of vibration (sec)', fontsize = 10)
    plt.ylabel('Spectral acceleration (g)', fontsize = 10)
    plt.subplot(1, 3, 2)
    plt.xlabel('Periods of vibration (sec)', fontsize = 10)
    plt.ylabel('Spectral displacement (m)', fontsize = 10)
    plt.subplot(1, 3, 3)
    plt.xlabel('Spectral displacement (m)', fontsize = 10)
    plt.ylabel('Spectral acceleration (g)', fontsize = 10)
    plt.show()

def NewmarkBeta(time, acc, T, damping):

    u0 = 0
    v0 = 0
    dt = time[1]-time[0]
    no_acc = len(acc)
    no_T = len(T)
    M = 1
    Sd = np.zeros(no_T)
    Sa = np.zeros(no_T)
    u = np.zeros(no_acc)
    a = np.zeros(no_acc)
    v = np.zeros(no_acc)
    at = np.zeros(no_acc)

    for i in range(no_T):
        if T[i] == 0:
            Sd[i] = 0
            Sa[i] = max(abs(np.array(acc)))
        else:
            wn = 2*math.pi/T[i]
            C = damping*2*M*wn
            K = ((2*math.pi)/T[i])**2*M
            u[0] = u0
            v[0] = v0
            a[0] = -acc[0]-C*v[0]-K*u[0]
            at[0] = acc[0]+a[0]
            for j in range(no_acc-1):
                u[j+1] = u[j] + dt*v[j] + dt**2 / 2*a[j]
                a[j+1] = (1/(M+dt*0.5*C)) * (-M*acc[j+1] - K*u[j+1] - C*(v[j]+dt*0.5*a[j]))
                v[j+1] = v[j] + dt*(0.5*a[j] + 0.5*a[j+1])
                at[j+1] = acc[j+1] + a[j+1]

            Sd[i] = max(abs(u))
            Sa[i] = max(abs(at))/9.81

    spectrum = {'Sd': None, 'Sa': None}
    spectrum['Sd'] = Sd
    spectrum['Sa'] = Sa

    return spectrum

def isNumber(s):
    try:
        float(s)
        return True
    except ValueError:
        return False

def read_damage_model(input_file):

    damage_states = []
    type_damage_state = []
    distribution = []
    mean = []
    cov = []
    median = []
    dispersion = []

    with open(input_file) as f:
        data = f.readlines()
        line = data[0]
        line = line.strip().split(',')
        type_criteria = line[1]

    with open(input_file, 'rU') as f:
        data = [row for row in csv.reader(f)]
        if (type_criteria == 'interstorey drift'):
            if data[2][1]== 'TRUE':
                path_to_deformed_shape = data[2][2]
                index = input_file.rfind('/')
                entire_path_to_deformed_shape = input_file[0:index+1]+path_to_deformed_shape
            else:
                entire_path_to_deformed_shape = 'none'

            damage_states = [row[0] for row in data[3:]]
            distribution = [row[1] for row in data[3:]]
            no_capacity = (len(data[3])-2)/2
            for icc in range(0, no_capacity):
                submedian = []
                subdispersion = []
                for iline in data[3:]:
                    submedian.append(float(iline[icc*2+2]))
                    subdispersion.append(float(iline[icc*2+3]))
                median.append(submedian)
                dispersion.append(subdispersion)

            damage_model = {'type_criteria': None, 'damage_states': None,
                 'median': None, 'dispersion': None}
            damage_model['type_criteria'] = type_criteria
            damage_model['damage_states'] = damage_states
            damage_model['distribution'] = distribution
            damage_model['median'] = median
            damage_model['dispersion'] = dispersion
            damage_model['deformed shape'] = entire_path_to_deformed_shape

        else:
            for iline in range(len(data)-2):
                line = data[iline+2]
                damage_states.append(line[0])
                if type_criteria == 'capacity curve dependent':
                    print(line)
                    type_damage_state.append(line[1])
                    distribution.append(line[2])
                    if isNumber(line[3]):
                        mean.append(float(line[3]))
                    else:
                        mean.append(line[4])
                    cov.append(float(line[4]))

                if type_criteria == 'strain dependent':
                    type_damage_state.append(line[1])
                    distribution.append(line[2])
                    mean.append(extract_values_string(line[3]))
                    cov.append(extract_values_string(line[4]))

                if type_criteria == 'spectral displacement':
                    distribution.append(line[1])
                    no_capacity = (len(line)-2)/2
                    mean.append(float(line[2]))
                    cov.append(float(line[3]))

            damage_model = {'type_criteria': None, 'damage_states': None,
                'type_damage_state': None, 'mean': None, 'cov': None}
            damage_model['type_criteria'] = type_criteria
            damage_model['damage_states'] = damage_states
            damage_model['type_damage_state'] = type_damage_state
            damage_model['distribution'] = distribution
            damage_model['mean'] = mean
            damage_model['cov'] = cov

    return damage_model

def extract_values_string(string):

    string = string.split(' ')
    values = []
    if isNumber(string[0]):
        for value in string:
            values.append(float(value))

    return values

def define_limit_states(capacity_curves, icc, damage_model):

    no_damage_states = len(damage_model['damage_states'])
    limit_states = []
    assert(damage_model['type_criteria'] != 'strain dependent'), 'Strain dependent damage model cannot be used in this methodology'
    if damage_model['type_criteria'] == 'capacity curve dependent':
        Sd = capacity_curves['Sd'][icc]
        Sa = capacity_curves['Sa'][icc]
        assert(capacity_curves['Sdy'] != []), 'If you want to use capacity curve dependent damage model you should provide Sdy, Say in capacity curve input file'
        Sdy = capacity_curves['Sdy'][icc]
        Say = capacity_curves['Say'][icc]
        for ids in range(no_damage_states):
            type_damage_state = damage_model['type_damage_state'][ids]
            distribution = damage_model['distribution'][ids]
            mean = damage_model['mean'][ids]
            cov = damage_model['cov'][ids]
            limit_states.append(define_limit_state(Sd, Sa, Sdy, Say,
                                                   type_damage_state,
                                                   distribution, mean, cov))

    elif damage_model['type_criteria'] == 'spectral displacement':
        for ids in range(no_damage_states):
            mean = damage_model['mean'][ids]
            cov = damage_model['cov'][ids]
            limit_states.append(sample_value(damage_model['distribution'][ids], mean, cov,0, float('inf')))

    elif damage_model['type_criteria'] == 'interstorey drift':
        if len(damage_model['median'])==1:
            EDPlim = damage_model['median'][0]
            bUthd = damage_model['dispersion'][0]
        else:
            EDPlim = damage_model['median'][icc]
            bUthd = damage_model['dispersion'][icc]

        [ISDvec, Sdvec] = read_deformed_shape(damage_model, capacity_curves,icc)
        [ISDvec,indy] = np.unique(ISDvec,return_index=True);
        Sdvec=Sdvec[indy];
        duf = capacity_curves['Sd'][icc][-1]

        for ids in range(no_damage_states):
            if bUthd[ids]==0: EDPsample = EDPlim[ids]
            else: EDPsample = stats.lognorm.rvs(bUthd[ids], scale = EDPlim[ids])
            Sdlim = np.interp(EDPsample,ISDvec,Sdvec);
            if Sdlim>duf: Sdlim = duf
            limit_states.append(Sdlim)

    #print(limit_states)
    return limit_states

def define_limit_state(Sd, Sa, Sdy, Say, type_damage_state, distribution, mean, cov):
    #FIXME: calls to the function sample_value are missing arguments A and B
    if type_damage_state == 'Sdy':
        mean = Sdy
        limit_state = sample_value(distribution, mean, cov, 0, float('inf'))

    elif type_damage_state == 'Sdu':
        mean = max(Sd)
        limit_state = sample_value(distribution, mean, cov, 0, float('inf'))

    elif type_damage_state == 'fraction Sdy':
        mean = mean*Sdy
        limit_state = sample_value(distribution, mean, cov, 0, float('inf'))

    elif type_damage_state == 'fraction Sdu':
        #FIXME: Sdu is not defined
        mean = mean*Sdu
        limit_state = sample_value(distribution, mean, cov, 0, float('inf'))

    elif type_damage_state == 'mean Sdy Sdu':
        mean = (Sdy+max(Sd))/2
        limit_state = sample_value(distribution, mean, cov, 0, float('inf'))

    elif isNumber(type_damage_state[0]):
        values = type_damage_state.split()
        weight1 = float(values[0])
        weight2 = float(values[2])
        mean = (weight1*Sdy + weight2*max(Sd))/(weight1+weight2)
        limit_state = sample_value(distribution, mean, cov, 0, float('inf'))

    return limit_state

def sample_value(distribution, mean, cov, A, B):

    if cov == 0:
        result = mean
    else:
        result = float('-inf')

        while result <=  A or result > B:
            if distribution == 'normal':
                result = stats.norm.rvs(mean, mean*cov)
            elif distribution == 'lognormal':
                variance = (mean*cov)**2.0
                mu = math.log(mean ** 2.0 / math.sqrt(variance + mean ** 2.0) )
                sigma = math.sqrt(math.log((variance / mean ** 2.0) + 1.0))
                result = stats.lognorm.rvs(sigma, scale = math.exp(mu))
            elif distribution == 'gamma':
                beta = (mean*cov)**2/mean
                alpha = mean/beta
                result = stats.gamma.rvs(alpha, scale = beta)

    return result


def spread(array, no_steps):

    result = np.zeros((len(array)-1)*no_steps+1)

    for i in range(len(array)-1):
        result[i*no_steps] = array[i]
        step = (array[i+1] - array[i]) / no_steps
        for j in range(no_steps):
            result[i*no_steps + j + 1] = array[i] + (j+1)*step

    result[-1] = array[-1]

    return result

def allocate_damage(igmr, PDM, disp, limitStates):

    no_ls = len(limitStates)
    #print(PDM[igmr, :])
    PDM[igmr, 0] = PDM[igmr, 0]+1
    ds = 0
    for ils in range(no_ls):
        if disp > limitStates[no_ls - ils - 1]:
            ds = no_ls-ils
            PDM[igmr, ds] = PDM[igmr, ds]+1
            PDM[igmr, 0] = PDM[igmr, 0]-1
            break
            
    #print(disp)
    #print(PDM[igmr, :])
    return PDM, ds

def residuals(coeffs, y, x):
    # TODO: this function seems incomplete - x not used, IMLs not defined
    res = y - stats.lognorm.cdf(IMLs, coeffs[1], scale = math.exp(coeffs[0]))

    return res

def calculate_imls(gmrs, T, damping):

    Sa = []
    Sd = []
    pga = []
    HI = []

    for igmr in range(len(gmrs['time'])):
        time = gmrs['time'][igmr]
        acc = gmrs['acc'][igmr]
        try:
            spectrum = NigamJennings(time, acc, [T], damping)
            Sa.append(spectrum['Sa'][0])
            Sd.append(spectrum['Sd'][0])
            pga.append(max(abs(np.array(acc)))/9.81)
        except (ValueError):
            spectrum = NigamJennings(time, acc, T, damping)
            dT = T[1]- T[0]
            HI.append(sum(spectrum['PSv'])*dT)

    return Sa, Sd, pga, HI

def calculate_iml(igmr, gmrs, IMT, T, damping):

    time = gmrs['time'][igmr]
    acc = gmrs['acc'][igmr]
    
    try:
        spectrum = NigamJennings(time, acc, [T], damping)
    except (ValueError):
        spectrum = NigamJennings(time, acc, T, damping)

    if IMT == 'PGA':
        iml = max(abs(np.array(acc)))
    elif IMT == 'Sa':
        if len(spectrum['Sa'])== 1: iml = spectrum['Sa'][0]
        else: iml = spectrum['Sa']
    elif IMT == 'Sd':
        if len(spectrum['Sd'])== 1: iml = spectrum['Sd'][0]
        else: iml = spectrum['Sd']
    elif IMT == 'Sv':
        if len(spectrum['Sv'])== 1: iml = spectrum['Sv'][0]
        else: iml = spectrum['Sv']
    elif IMT == 'PSv':
        if len(spectrum['PSv'])== 1: iml = spectrum['PSv'][0]
        else: iml = spectrum['PSv']
    elif IMT == 'HI':
        # computing Housner Intensity
        dT = T[1]- T[0]
        iml = sum(spectrum['PSv'])*dT;

    return iml

def calculate_mean_fragility(gmrs, EDPs, T, damping, IMT, damage_model, method):

    imls = calculate_imls(gmrs, T, damping)

    if IMT == 'Sa':
        imls = imls[0]
        IMT = 'Sa('+str(T)+')'
    elif IMT == 'Sd':
        imls = imls[1]
        IMT = 'Sd('+str(T)+')'
    elif IMT == 'PGA':
        imls = imls[2]

    if method == 'cloud analysis':
        Sds = EDPs[1]
        #calculate log(SDs) versus log(IMLs)
        #regression analysis using censored
        #with 
        
    else:
        PDM = response[0]
        cumPDM = convert_damage_matrix_to_probability(PDM)

        logmeans, logstddev = [], []
        if method == 'least squares':
            for iDS in range(len(damage_model['damage_states'])):
                solution, _ = optimize.leastsq(residual_lognormal_dist,[0.1, 0.6],args=(imls, cumPDM[iDS+1]))
                logmeans.append(solution[0])
                logstddev.append(solution[1])

        elif method == 'max likelihood':
            # the imls and the PDM are sorted to be used in the maximum likelihood approach
            # the PDM should be a list of integers in this approach
            I = np.argsort(imls, axis=0)
            imls = np.array(imls)
            imls = imls[I]
            PDM_mle = []
            for iDS in range(len(damage_model['damage_states'])):
                temp = cumPDM[iDS+1][I]*no_assets
                temp2 = temp.astype(int)
                PDM_mle = temp2.tolist()
                solution = mle(PDM_mle,imls,no_assets,1)
                logmeans.append(solution[0])
                logstddev.append(solution[1])
            
        Rsquare = calculate_correlation(logmeans,logstddev,cumPDM,imls)


    fragility_model = {'damage_states': None, 'logmean': None,'logstddev': None, 'IMT': None}
    fragility_model['damage_states'] = damage_model['damage_states']
    fragility_model['logmean'] = logmeans
    fragility_model['logstddev'] = logstddev
    fragility_model['IMT'] = IMT
    fragility_model['Rsquare'] = Rsquare

    return fragility_model

def mle(num_collapse, iml, tot, g):
#This function gives maximum likelihood estimate of a lognormal function

    def func(x):
        p = [stats.lognorm.cdf(i, x[0], loc=0, scale=x[1]) for i in iml]
        return -np.sum(stats.binom.logpmf(num_collapse,tot,p))
    x0 = np.array([1, g])
    x = optimize.fmin(func, x0)
    sigma = x[0]
    mu = np.log(x[1])

    return [mu, sigma]

def save_mean_fragility(taxonomy, fragility_model, minIML, maxIML, output_type, output_path):

    damage_states = fragility_model['damage_states']
    logmeans = fragility_model['logmean']
    logstddev = fragility_model['logstddev']
    Rsquares = fragility_model['Rsquare']
    IMT = fragility_model['IMT']

    if output_type == 'csv':
        save_mean_fragility_csv(taxonomy, damage_states, logmeans, logstddev, Rsquares, minIML, maxIML, IMT, output_path)
    if output_type == 'nrml':
        save_mean_fragility_nrml(taxonomy, damage_states, logmeans, logstddev, minIML, maxIML, IMT, output_path)

def save_mean_fragility_csv(taxonomy, damage_states, logmeans, logstddev, Rsquares, minIML, maxIML, IMT, output_path):

    destination_file = os.path.join(output_path, taxonomy)
    with open(output_path+'/'+taxonomy+'_fra.csv', 'wb') as csvfile:
        writer = csv.writer(csvfile)
        writer.writerow([taxonomy, IMT, minIML, maxIML])
        writer.writerow(['Damage state', 'log mean', 'log stddev', 'mean', 'stddev', 'median', 'cov','Rsquare'])
        for iDS in range(len(damage_states)):
            mu = logmeans[iDS]
            sigma = logstddev[iDS]
            mean = math.exp(mu+sigma**2/2)
            stddev = math.sqrt(math.exp(2*mu+sigma**2)*(math.exp(sigma**2)-1))
            median = math.exp(mu)
            cov = abs(stddev/mean)
            Rsquare = Rsquares[iDS]
            writer.writerow([damage_states[iDS], mu, sigma, mean, stddev, median, cov, Rsquare])

def get_damage_states(csv_data):
    damage_states = []
    csv_data.next()
    csv_data.next()
    for line in csv_data:
        damage_states.append(line[0])
    return damage_states

def save_fragility_set_nrml(folder, destination_file):
    path = os.path.abspath(folder)
    destination_file = os.path.join(path, destination_file)
    nrml_file = open(destination_file, 'w')
    nrml_file.write('<?xml version="1.0" encoding="UTF-8"?>\n')
    nrml_file.write('<nrml xmlns="http://openquake.org/xmlns/nrml/0.5">\n\n')
    nrml_file.write('<fragilityModel id="fm" assetCategory="buildings" lossCategory="structural">\n\n')
    nrml_file.write('  <description>fragility model</description>\n')
    ds_str = '  <limitStates>'

    file_index = 0
    for f in os.listdir(folder):
        if f.endswith(".csv"):
            with open(os.path.join(path, f), 'rU') as fr:
                data = csv.reader(fr)
                if file_index == 0:
                    damage_states = get_damage_states(data)
                    fr.seek(0)
                    ds_str += ' '.join(damage_states)
                    nrml_file.write(ds_str + '</limitStates>\n\n')
                    file_index += 1
                taxonomy, IMT, minIML, maxIML = data.next()[0:4]
                data.next()

                if IMT == 'PGA' or IMT[:2] == 'Sa':
                    IMT_units = 'g'
                else:
                    IMT_units = 'm'

                nrml_file.write('  <fragilityFunction format="continuous" id="' + taxonomy + '">\n')
                nrml_file.write('    <imls imt="'+IMT+'" minIML="'+minIML+'" maxIML="'+maxIML+'"/>\n')

                for line in data:
                    ds, logmean, logstd, mean, stddev, median, cov = line[0:7]
                    nrml_file.write('    <params ls="'+ds+'" mean="'+mean+'" stddev="'+stddev+'"/>\n')
                nrml_file.write('  </fragilityFunction>\n\n')

    nrml_file.write('</fragilityModel>\n\n')
    nrml_file.write('</nrml>\n')
    nrml_file.close()

def save_mean_fragility_nrml(taxonomy, damage_states, logmeans, logstddev, minIML, maxIML, IMT, output_path):

    path = os.path.abspath(output_path)
    destination_file = os.path.join(path, taxonomy)
    nrml_file = open(destination_file+'_fra.xml', 'w')
    nrml_file.write('<?xml version = "1.0" encoding = "UTF-8"?>\n')
    nrml_file.write('<nrml xmlns = "http://openquake.org/xmlns/nrml/0.4">\n')
    nrml_file.write('<fragilityModel format = "continuous">\n')
    nrml_file.write('   <description>fragility model for '+ taxonomy +'</description>\n')
    ds_str = ' <limitStates>'
    for damage_state in damage_states:
        ds_str = ds_str + ' ' + damage_state
    nrml_file.write(ds_str + '</limitStates>\n')
    nrml_file.close()

    save_single_fragility_nrml(taxonomy, damage_states, logmeans, logstddev, minIML, maxIML, IMT)

    nrml_file = open(taxonomy+'_fra.xml', 'a')
    nrml_file.write('   </fragilityModel>\n')
    nrml_file.write('</nrml>\n')
    nrml_file.close()

def save_single_fragility_nrml(taxonomy, damage_states, logmeans, logstddev, minIML, maxIML, IMT):

    if IMT == 'PGA' or IMT[:2] == 'Sa':
        IMT_units = 'g'
    else:
        IMT_units = 'm'

    nrml_file = open(taxonomy+'_fra.xml', 'a')
    nrml_file.write('       <ffs type = "lognormal">\n')
    nrml_file.write('       <taxonomy>'+taxonomy+'</taxonomy>\n')
    nrml_file.write('       <IML IMT = "'+IMT+'" imlUnit = "'+IMT_units+'" minIML = "'+str(minIML)+'" maxIML = "'+str(maxIML)+'"/>\n')

    for iDS in range(len(damage_states)):
        mu = logmeans[iDS]
        sigma = logstddev[iDS]
        mean = math.exp(mu+sigma**2/2)
        stddev = math.sqrt(math.exp(2*mu+sigma**2)*(math.exp(sigma**2)-1))
        nrml_file.write('       <ffc ls = "'+damage_states[iDS]+'">\n')
        nrml_file.write('           <params mean = "'+str(mean)+'" stddev = "'+str(stddev)+'"/>\n')
        nrml_file.write('       </ffc>\n')
    nrml_file.write('       </ffs>\n')
    nrml_file.close()

def plot_fragility_model(fragility_model, minIML, maxIML):

    imls = np.linspace(minIML, maxIML, 100)
    color_scheme = ['g', 'b', 'y', 'orangered', 'r', 'k', 'm', 'c', '0.5', '0.75']
    for iDS in range(len(fragility_model['damage_states'])):
        mu = fragility_model['logmean'][iDS]
        sigma = fragility_model['logstddev'][iDS]
        if sigma > 0:
            plt.plot(imls, stats.lognorm.cdf(imls, sigma,
                                         scale=math.exp(mu)),
                                         color=color_scheme[iDS],
                                         linewidth=2,
                                         label= fragility_model['damage_states'][iDS])
        else: #sigma = 0
            PoE = []
            for ele in imls:
                if ele <=math.exp(mu): PoE.append(0)
                else: PoE.append(1)

            plt.plot(
                imls,PoE,color=color_scheme[iDS],
                linewidth=2,label= fragility_model['damage_states'][iDS])


        try:
            plt.plot(fragility_model['imls'],fragility_model['PoE points'][iDS+1],
                                    marker='o',color=color_scheme[iDS],linestyle='None')
        except:
            pass
        plt.xlabel('Sa(Tel) [g]', fontsize = 12)
        plt.ylabel('Probabilty of Exceedance', fontsize = 12)
        plt.legend(loc=4)
    plt.show()
    
    
def convert_damage_matrix_to_probability(PDM):
    
    no_assets = np.sum(PDM, axis=1, dtype=float)
    cumPDM = np.fliplr(np.cumsum(np.fliplr(PDM), axis=1))
    cumPDM = np.transpose(cumPDM)
    cumPDM = cumPDM/no_assets
    
    return cumPDM
        
def export_IMLs_PDM(gmrs,T,PDM,damping_ratio,damage_model,output_file):
    
    cumPDM = convert_damage_matrix_to_probability(PDM)
    Sa, Sd, PGA, HI = calculate_imls(gmrs, T, damping_ratio)
    header = []
    header.append('Sa('+str(T)+')')
    header.append('Sd('+str(T)+')')
    header.append('PGA')
    for iDS in range(len(damage_model['damage_states'])):
        header.append(damage_model['damage_states'][iDS])
            
    with open(output_file, 'wb') as csvfile:
        writer = csv.writer(csvfile)
        writer.writerow(header)
        for irow in range(len(Sa)):
            row = []
            row.append(Sa[irow])
            row.append(Sd[irow])
            row.append(PGA[irow])
            for iDS in range(len(damage_model['damage_states'])):
                row.append(cumPDM[iDS+1][irow])
            writer.writerow(row)

def plot_fragility_scatter(fragility_model, minIML, maxIML, PDM, gmrs, IMT, T, damping_ratio):

    imls = np.linspace(minIML, maxIML, 100)
    Sa, Sd, PGA, HI = calculate_imls(gmrs, T, damping_ratio)
    cumPDM = convert_damage_matrix_to_probability(PDM)
    color_scheme = ['g', 'b', 'y', 'orangered', 'r', 'k', 'm', 'c', '0.5', '0.75']
    
    for iDS in range(len(fragility_model['damage_states'])):
        mu = fragility_model['logmean'][iDS]
        sigma = fragility_model['logstddev'][iDS]
        plt.scatter(Sa, cumPDM[iDS+1], s=20, c=color_scheme[iDS], alpha=0.5)
        if sigma > 0:
            plt.plot(imls, stats.lognorm.cdf(imls, sigma,
                                         scale=math.exp(mu)),
                                         color=color_scheme[iDS],
                                         linewidth=2,
                                         label= fragility_model['damage_states'][iDS])
        else: #sigma = 0
            PoE = []
            for ele in imls:
                if ele <=math.exp(mu): PoE.append(0)
                else: PoE.append(1)

            plt.plot(
                imls,PoE,color=color_scheme[iDS],
                linewidth=2,label= fragility_model['damage_states'][iDS])


        try:
            plt.plot(fragility_model['imls'],fragility_model['PoE points'][iDS+1],
                                    marker='o',color=color_scheme[iDS],linestyle='None')
        except:
            pass
        plt.xlabel('Sa(Tel) [g]', fontsize = 12)
        plt.ylabel('Probabilty of Exceedance', fontsize = 12)
        plt.legend(loc=4)
        plt.ylim((0,1))
        plt.xlim((minIML,maxIML))
    plt.show()
    
def save_result(result,output_file):
    
    np.savetxt(output_file,result,delimiter=',')
    
def import_result(input_file):
    
    result = np.genfromtxt(input_file,delimiter=',')
    
    return result 


def NigamJennings(time, acc, periods, damping):

    add_PGA = False
    if periods[0] == 0:
        periods = np.delete(periods, 0)
        add_PGA = True

    dt = time[1]-time[0]
    num_steps = len(acc)
    num_per = len(periods)
    vel, disp = calculate_velocity_displacement(time, acc)
    omega = (2. * np.pi) / np.array(periods)
    omega2 = omega ** 2.
    omega3 = omega ** 3.
    omega_d = omega * math.sqrt(1.0 - (damping ** 2.))
    const = {'f1': (2.0 * damping) / (omega3 * dt),
            'f2': 1.0 / omega2,
            'f3': damping * omega,
            'f4': 1.0 / omega_d}
    const['f5'] = const['f3'] * const['f4']
    const['f6'] = 2.0 * const['f3']
    const['e'] = np.exp(-const['f3'] * dt)
    const['s'] = np.sin(omega_d * dt)
    const['c'] = np.cos(omega_d * dt)
    const['g1'] = const['e'] * const['s']
    const['g2'] = const['e'] * const['c']
    const['h1'] = (omega_d * const['g2']) - (const['f3'] * const['g1'])
    const['h2'] = (omega_d * const['g1']) + (const['f3'] * const['g2'])
    x_a, x_v, x_d = calculate_time_series(num_steps, num_per, time, acc, const, omega2)

    spectrum = {'Sa': None, 'Sv': None, 'Sd': None, 'T': None}
    spectrum['Sa'] = np.max(np.fabs(x_a)/9.81, axis = 0)
    spectrum['Sv'] = np.max(np.fabs(x_v), axis = 0)
    spectrum['Sd'] = np.max(np.fabs(x_d), axis = 0)
    spectrum['PSv'] = spectrum['Sd']*omega
    spectrum['T'] = periods

    if add_PGA:
        spectrum = add_PGA_spectrum(spectrum, acc)

    return spectrum

def add_PGA_spectrum(spectrum, acc):

    spectrum['Sa'] = np.append(np.max(np.fabs(acc)/9.81), spectrum['Sa'])
    spectrum['Sv'] = np.append(0, spectrum['Sv'])
    spectrum['Sd'] = np.append(0, spectrum['Sd'])
    spectrum['T'] = np.append(0, spectrum['T'])

    return spectrum

def calculate_time_series(num_steps, num_per, time, acc, const, omega2):
    """
    Calculates the acceleration, velocity and displacement time series for
    the SDOF oscillator
    :param dict const:
        Constants of the algorithm
    :param np.ndarray omega2:
        Square of the oscillator period
    :returns:
        x_a = Acceleration time series
        x_v = Velocity time series
        x_d = Displacement time series
    """
    dt = time[1]-time[0]
    x_d = np.zeros([num_steps - 1, num_per], dtype = float)
    x_v = np.zeros_like(x_d)
    x_a = np.zeros_like(x_d)

    for k in range(0, num_steps - 1):
        yval = k - 1
        dug = acc[k + 1] - acc[k]
        z_1 = const['f2'] * dug
        z_2 = const['f2'] * acc[k]
        z_3 = const['f1'] * dug
        z_4 = z_1 / dt
        if k == 0:
            b_val = z_2 - z_3
            a_val = (const['f5'] * b_val) + (const['f4'] * z_4)
        else:
            b_val = x_d[k - 1, :] + z_2 - z_3
            a_val = (const['f4'] * x_v[k - 1, :]) +\
                (const['f5'] * b_val) + (const['f4'] * z_4)

        x_d[k, :] = (a_val * const['g1']) + (b_val * const['g2']) +\
            z_3 - z_2 - z_1
        x_v[k, :] = (a_val * const['h1']) - (b_val * const['h2']) - z_4
        x_a[k, :] = (-const['f6'] * x_v[k, :]) - (omega2 * x_d[k, :])

    return x_a, x_v, x_d

def calculate_velocity_displacement(time, acc):
    '''
    Returns the velocity and displacment time series using simple integration
    :param float time_step:
        Time-series time-step (s)
    :param np.ndarray acceleration:
        Acceleration time-history
    :returns:
        velocity - Velocity Time series (cm/s)
        displacement - Displacement Time series (cm)
    '''
    dt = time[1]-time[0]
    velocity = dt * integrate.cumtrapz(acc, initial = 0.)
    displacement = dt * integrate.cumtrapz(velocity, initial = 0.)
    return velocity, displacement

def residual_lognormal_dist(coeffs, imls, fractions):

    mu = coeffs[0]
    sigma = coeffs[1]
    residual = abs(np.array(fractions) - stats.lognorm.cdf(imls, sigma, scale = math.exp(mu)))

    return residual


def calculate_fragility_statistics(gmrs, PDM, T, damping, IMT, damage_model,
                                   method, no_samples, size_sample):

    imls = calculate_imls(gmrs, T, damping)

    if IMT == 'Sa':
        imls = imls[0]
        IMT = 'Sa('+str(T)+')'
    elif IMT == 'Sd':
        imls = imls[1]
        IMT = 'Sd('+str(T)+')'
    elif IMT == 'PGA':
        imls = imls[2]

    no_gmrs = len(imls)
    no_DS = len(damage_model['damage_states'])
    cumPDM = convert_damage_matrix_to_probability(PDM)

    statistics = []
    for isample in range(no_samples):
        sample_indices = np.random.random_integers(0, no_gmrs-1, size_sample)
        sample_imls = imls[sample_indices]
        sample_cumPDM = cumPDM[:, sample_indices]
        fragility_model = []
        if method == 'least squares':
            for iDS in range(no_DS):
                solution, _ = optimize.leastsq(residual_lognormal_dist,
                                               [0.1, 0.6],
                                               args=(sample_imls, sample_cumPDM[iDS+1]))
                fragility_model.append(solution)
        elif method == 'max likelihood':
            I = np.argsort(sample_imls, axis=0)
            imls = np.array(sample_imls)
            imls = sample_imls[I]
            sample_PDM_mle = []
            for iDS in range(no_DS):
                temp = sample_cumPDM[iDS+1][I]*no_assets
                temp2 = temp.astype(int)
                sample_PDM_mle = temp2.tolist()
                solution = mle(sample_PDM_mle, sample_imls, no_assets, 1)
                fragility_model.append(solution)
        statistics.append(fragility_model)

    statistics = np.array(statistics)
    means = np.mean(statistics, axis=0)
    stddev = np.std(statistics, axis=0)

    fragility_stats = {'damage_states': None, 'mean': None, 'stddev': None,
        'correlation': None, 'IMT': None}
    fragility_stats['damage_states'] = damage_model['damage_states']
    fragility_stats['mean'] = np.reshape(means, (no_DS, 2))
    fragility_stats['stddev'] = np.reshape(stddev, (no_DS, 2))
    fragility_stats['correlation'] = np.corrcoef(np.transpose(statistics))
    fragility_stats['statistics'] = statistics
    fragility_stats['IMT'] = IMT

    return fragility_stats

def plot_fragility_stats(fragility_stats, minIML, maxIML):

    imls = np.linspace(minIML, maxIML, 100)
    color_scheme = ['g', 'b', 'y', 'orangered', 'r', 'k', 'm', 'c', '0.5', '0.75']
    for iDS in range(len(fragility_stats['damage_states'])):
        mu = fragility_stats['mean'][iDS][0]
        sigma = fragility_stats['mean'][iDS][1]
        plt.plot(imls, stats.lognorm.cdf(imls, sigma, scale = math.exp(mu)),
                 color = color_scheme[iDS],  linewidth = 2,
                 label = fragility_stats['damage_states'][iDS])
        plt.legend(loc = 4)
    plt.show()

def read_consequence_model(input_file):

    damage_states, distribution, mean, cov, A, B = [], [], [], [], [], []
    with open(input_file, 'rU') as f:
        data = csv.reader(f)
        for iline, row in enumerate(data):
            if iline > 0:
                damage_states.append(row[0])
                distribution.append(row[1])
                mean.append(float(row[2]))
                cov.append(float(row[3]))
                A.append(float(row[4]))
                B.append(float(row[5]))

    cons_model = {'damage_states': None, 'distribution': None, 'mean': None,
        'cov': None, 'A': None, 'B': None}
    cons_model['damage_states'] = damage_states
    cons_model['distribution'] = distribution
    cons_model['mean'] = mean
    cons_model['cov'] = cov
    cons_model['A'] = A
    cons_model['B'] = B

    return cons_model

def convert_fragility_vulnerability(fragility_model, cons_model, imls, dist_type):

    vulnerability_model = []
    check_damage_states(fragility_model, cons_model)
    no_samples = 100
    loss_ratios = []
    for isample in range(no_samples):
        loss_ratios.append(sample_loss_ratios(fragility_model, cons_model, imls))
    if dist_type == 'lognormal' or dist_type == 'beta':
        vulnerability_model = create_parametric_vul_model(imls, loss_ratios, dist_type, fragility_model['IMT'])
    elif dist_type == 'PMF':
        vulnerability_model = create_nonparametric_vul_model(imls, loss_ratios, dist_type, fragility_model['IMT'])

    return vulnerability_model

def create_parametric_vul_model(imls, loss_ratios, dist_type, IMT):

    mean = np.mean(np.array(loss_ratios), axis=0)
    stddev = np.std(np.array(loss_ratios), axis=0)
    cov = np.zeros(len(mean))
    for icov in range(len(cov)):
        if mean[icov] > 0:
            if stddev[icov]/mean[icov] > 10**-4:
                cov[icov] = stddev[icov]/mean[icov]

    vulnerability_model = {'imls': None, 'distribution': None,
        'mean': None, 'cov': None, 'IMT': None}
    vulnerability_model['imls'] = imls
    vulnerability_model['distribution'] = dist_type
    vulnerability_model['mean'] = mean
    vulnerability_model['cov'] = cov
    vulnerability_model['IMT'] = IMT

    return vulnerability_model

def create_nonparametric_vul_model(imls, sampled_loss_ratios, dist_type, IMT):

    no_values = 20
    no_samples = len(sampled_loss_ratios)
    loss_ratios = np.linspace(0.0, 1.0, no_values)
    probs = np.zeros((no_values, len(imls)))

    for isample in range(no_samples):
        for iiml in range(len(imls)):
            if sampled_loss_ratios[isample][iiml] == 0:
                probs[0][iiml] = probs[0][iiml] + 1
            elif sampled_loss_ratios[isample][iiml] == 1:
                probs[-1][iiml] = probs[-1][iiml] + 1
            else:
                idx = (np.abs(loss_ratios[1:-1]-sampled_loss_ratios[isample][iiml])).argmin()
                probs[idx+1][iiml] =  probs[idx+1][iiml] + 1

    probs = probs/no_samples
    vulnerability_model = {'imls': None, 'distribution': None,
        'loss_ratios': None, 'probabilities': None, 'IMT': None}
    vulnerability_model['imls'] = imls
    vulnerability_model['distribution'] = dist_type
    vulnerability_model['loss_ratios'] = loss_ratios
    vulnerability_model['probabilities'] = probs
    vulnerability_model['IMT'] = IMT

    return vulnerability_model

def plot_vulnerability_model(vulnerability_model):
    imls = vulnerability_model['imls']
    imt = vulnerability_model['IMT']
    if vulnerability_model['distribution'] == 'lognormal':
        mean_lrs = vulnerability_model['mean']
        cov_lrs = vulnerability_model['cov']
        plt.errorbar(imls, mean_lrs, yerr=cov_lrs, color='0.4', fmt='o')
    elif vulnerability_model['distribution'] == 'beta':
        mean_lrs = vulnerability_model['mean']
        cov_lrs = vulnerability_model['cov']
        plt.errorbar(imls, mean_lrs, yerr=cov_lrs, color='0.4', fmt='o')
    elif vulnerability_model['distribution'] == 'PMF':
        lrs = vulnerability_model['loss_ratios']
        probs = vulnerability_model['probabilities']
        mean_lrs = np.dot(lrs, probs)
        imls.insert(0, 0)
        mean_lrs = np.insert(mean_lrs, 0, 0)
    plt.plot(imls, mean_lrs, linewidth=2)
    plt.plot(imls, mean_lrs, 'ro')
    plt.ylim(0.0, 1.0)
    plt.xlabel(imt, fontsize=12)
    plt.ylabel('Mean Loss Ratio', fontsize=12)
    plt.suptitle('Vulnerability function', fontsize=16)
    plt.show()

def check_damage_states(model1, model2):

    for iDS in range(len(model1['damage_states'])):
        assert(model1['damage_states'][iDS] ==  model2['damage_states'][iDS]), 'The fragility model and the consequence model are not compatible'

    return

def sample_loss_ratios(fragility_model, cons_model, imls):

    no_damage_states = len(fragility_model['damage_states'])
    loss_ratios = np.zeros(len(imls))

    for iDS in range(no_damage_states):
        distribution = cons_model['distribution'][iDS]
        mean = cons_model['mean'][iDS]
        cov = cons_model['cov'][iDS]
        A = cons_model['A'][iDS]
        B = cons_model['B'][iDS]
        damage_ratio = sample_value(distribution, mean, cov, A, B)
        mu = fragility_model['logmean'][iDS]
        sigma = fragility_model['logstddev'][iDS]

        if iDS == no_damage_states-1:
            fraction = stats.lognorm.cdf(imls, sigma, scale = math.exp(mu))
        else:
            mu2 = fragility_model['logmean'][iDS + 1]
            sigma2 = fragility_model['logstddev'][iDS + 1]
            fraction = stats.lognorm.cdf(imls, sigma, scale = math.exp(mu))-stats.lognorm.cdf(imls, sigma2, scale = math.exp(mu2))
        loss_ratios = loss_ratios + fraction*damage_ratio

    return loss_ratios

def save_vulnerability(taxonomy, vulnerability_model, output_type, output_path):

    imls = vulnerability_model['imls']
    IMT = vulnerability_model['IMT']
    distribution = vulnerability_model['distribution']
    if distribution == 'lognormal' or distribution == 'beta':
        values1 = vulnerability_model['mean']
        values2 = vulnerability_model['cov']
    elif distribution == 'PMF':
        values1 = vulnerability_model['loss_ratios']
        values2 = vulnerability_model['probabilities']

    if output_type == 'csv':
        save_vulnerability_csv(taxonomy, imls, values1, values2, distribution, IMT, output_path)
    if output_type == 'nrml':
        save_vulnerability_nrml(taxonomy, imls, values1, values2, distribution, IMT, output_path)

def save_vulnerability_csv(taxonomy, imls, values1, values2, distribution, IMT, output_path):

    path = os.path.abspath(output_path)
    destination_file = os.path.join(path, taxonomy)
    with open(destination_file+'_vul.csv', 'wb') as csvfile:
        writer = csv.writer(csvfile)
        writer.writerow([taxonomy, IMT, distribution])
        if distribution == 'lognormal' or distribution == 'beta':
            writer.writerow(np.append('imls', imls))
            writer.writerow(np.append('mean', values1))
            writer.writerow(np.append('cov', values2))

        elif distribution == 'PMF':
            writer.writerow(['loss ratio', 'probabilities'])
            for iLR in range(len(values1)):
                writer.writerow(np.append(values1[iLR], values2[iLR, :]))


def save_vulnerability_nrml(taxonomy, imls, values1, values2, distribution, IMT, output_path):

    path = os.path.abspath(output_path)
    destination_file = os.path.join(path, taxonomy)
    nrml_file = open(destination_file+'_vul.xml', 'w')
    nrml_file.write('<?xml version = "1.0" encoding = "utf-8"?>\n')
    nrml_file.write('<nrml xmlns = "http://openquake.org/xmlns/nrml/0.5">\n')
    nrml_file.write('<vulnerabilityModel id = "'+taxonomy+'" assetCategory = "buildings" lossCategory = "economic">\n')
    nrml_file.write('   <description>vulnerability model for '+ taxonomy +'</description>\n')
    nrml_file.close()

    save_single_vulnerability_nrml(taxonomy, imls, values1, values2, distribution, IMT)

    nrml_file = open(taxonomy+'_vul.xml', 'a')
    nrml_file.write('</vulnerabilityModel>\n')
    nrml_file.write('</nrml>\n')
    nrml_file.close()

def save_single_vulnerability_nrml(taxonomy, imls, values1, values2, distribution, IMT):

    imls_str = convert_array_to_string(imls)
    nrml_file = open(taxonomy+'_vul.xml', 'a')

    if distribution == 'lognormal':
        dist = 'LN'
    elif distribution == 'beta':
        dist = 'BT'
    elif distribution == 'PMF':
        dist = 'PM'

    nrml_file.write('       <vulnerabilityFunction id = "'
                    + taxonomy + '" dist = "' + dist + '">\n')

    if distribution == 'lognormal' or distribution == 'beta':
        lrs_str = convert_array_to_string(values1)
        cov_str = convert_array_to_string(values2)
        nrml_file.write('       <imls imt = "'+IMT+'">'+imls_str+'</imls>\n')
        nrml_file.write('       <meanLRs>'+lrs_str+'</meanLRs>\n')
        nrml_file.write('       <covLRs>'+cov_str+'</covLRs>\n')

    elif distribution == 'PMF':
        for iLR in range(len(values1)):
            probs = convert_array_to_string(values2[iLR])
            nrml_file.write('       <probabilities lr = "'
                            + str(values1[iLR]) + '">' + probs
                            + '</probabilities>)\n')

    nrml_file.write('       </vulnerabilityFunction>\n')

def save_vulnerability_set_nrml(folder, destination_file):
    path = os.path.abspath(folder)
    destination_file = os.path.join(path, destination_file)
    nrml_file = open(destination_file, 'w')
    nrml_file.write('<?xml version="1.0" encoding="UTF-8"?>\n')
    nrml_file.write('<nrml xmlns="http://openquake.org/xmlns/nrml/0.5">\n\n')
    nrml_file.write('<vulnerabilityModel assetCategory="buildings" lossCategory="economic">\n')
    nrml_file.write('  <description>vulnerability model</description>\n\n')

    for f in os.listdir(folder):
        if f.endswith(".csv"):
            with open(os.path.join(path, f), 'rU') as fr:
                data = csv.reader(fr)
                taxonomy, IMT, distribution = data.next()[0:3]
                imls = data.next()
                imls.remove("imls")
                imls_str = " ".join(imls)
                if distribution == 'lognormal':
                    dist = 'LN'
                elif distribution == 'beta':
                    dist = 'BT'
                elif distribution == 'PMF':
                    dist = 'PM'

                nrml_file.write('  <vulnerabilityFunction id="'
                                + taxonomy + '" dist="' + dist + '">\n')
                nrml_file.write('    <imls imt="'+IMT+'">'+imls_str+'</imls>\n')

                if distribution == 'lognormal' or distribution == 'beta':
                    lrs = data.next()
                    lrs.remove("mean")
                    lrs_str = " ".join(lrs)
                    covs = data.next()
                    covs.remove("cov")
                    cov_str = " ".join(covs)
                    nrml_file.write('    <meanLRs>'+lrs_str+'</meanLRs>\n')
                    nrml_file.write('    <covLRs>'+cov_str+'</covLRs>\n')
                elif distribution == 'PMF':
                    data.next()
                    for line in data:
                        lr = line.pop(0)
                        probs = " ".join(line)
                        nrml_file.write('    <probabilities lr="' + lr + '">' + probs + '</probabilities>\n')

                nrml_file.write('  </vulnerabilityFunction>\n\n')

    nrml_file.write('</vulnerabilityModel>\n\n')
    nrml_file.write('</nrml>\n')
    nrml_file.close()

def convert_array_to_string(array):
    string = ''
    for i in range(len(array)):
        string = string + ' ' + str(array[i])

    return string

def read_hazard(input_file):
#This function reads a hazard curve and stores it in a dictionary

    file = open(input_file)
    data = csv.reader(file)
    IMLs = []
    prob_exceedance = []
    IM_type = []
    for line in data:
        if line[0] == 'PGA':
            IM_type.append(line[0])
            for value in line[1:]:
                if isNumber(value):
                    IMLs.append(float(value))
        if line[0] == 'Sa':
            IM_type.append(line[0])
            for value in line[1:]:
                if isNumber(value):
                    IMLs.append(float(value))
        if line[0] == 'PoE':
            for value in line[1:]:
                if isNumber(value):
                    prob_exceedance.append(float(value))

    #Store all the data in the dictionary
    hazard_curve = {'IMLs':None,'PoE':None,'IM_Type':None}
    hazard_curve['IMLs'] = IMLs
    hazard_curve['PoE'] = prob_exceedance
    hazard_curve['IM_Type'] = IM_type

    return hazard_curve

def plot_hazard_curve(hazard_curve):
#This function plots the hazard curve

    IMLs = hazard_curve['IMLs']
    PoE = hazard_curve['PoE']
    plt.plot(IMLs,PoE,color='g',linewidth=2)
    plt.xlabel(hazard_curve['IM_Type'][0] + ' [g]',fontsize = 10)
    plt.ylabel('Probability of Exceedance',fontsize = 10)

def read_frag_model(input_file):
#This function reads a fragility model and stores it in a dictionary

    damage_states = []
    model_type = []
    cent_value = []
    dispersion = []
    log_mean = []
    log_stdev = []

    file = open(input_file)
    data = file.readlines()
    line = data[0]
    line = line.strip().split(',')
    model_type = line[1] + ' - ' + line[2]

    file = open(input_file)
    data = csv.reader(file)
    data = [row for row in data]
    for iline in range(len(data)-1):
        line = data[iline+1]
        damage_states.append(line[0])
        cent_value.append(line[1])
        dispersion.append(line[2])

    if model_type == ' median -  dispersion':
        for ids in range(len(cent_value)):
            mean = math.log(cent.value[ids])
            stdev = dispersion[ids]
            log_mean.append(mean)
            log_stdev.append(stdev)
    elif model_type == 'mean of x - cov of x':
        for ids in range(len(cent_value)):
            mu = cent.value[ids]
            cov = dispersion[ids]
            median = (mu**2)/math.sqrt(mu**2+cov**2)
            mean = math.log(median)
            stdev = math.sqrt(math.log(1+(cov**2)/mu**2))
            log_mean.append(mean)
            log_stdev.append(stdev)
    elif model_type == 'mean of ln(x) - st. dev. of ln(x)':
        for ids in range(len(cent_value)):
            mean = cent_value[ids]
            stdev = line[ids]
            log_mean.append(mean)
            log_stdev.append(stdev)

    #Store all the data in the dictionary
    fragility_model =  {'Damage_states':None,'log_mean':None,'log_stdev':None}
    fragility_model['Damage_states'] = damage_states
    fragility_model['logmean'] = log_mean
    fragility_model['logstdev'] = log_stdev

    return fragility_model



"""
module to deal with lists, dictionaries and pandas

| dol: dict of lists
| lod: list of dicts
| pddf: pandas dataframe

    .. uml::

        class lists_and_dicts_and_pandas_etc <<(M,#FF7700)>> {
        .. functions ..
        + convert_lod_to_dol(list_of_dicts)
        + convert_dol_to_lod(dict_of_lists)
        + convert_lod_to_md_table(list_of_dicts)
        + convert_lod_to_pddf(list_of_dicts)
        + convert_dict_to_quantity_value(dict)
        }



"""

from tabulate import tabulate
import pandas as pd


def convert_lod_to_dol(list_of_dicts):
    """
    converts list of dicts --> dict of lists
    """
    _dol = {k: [dic[k] for dic in list_of_dicts] for k in list_of_dicts[0]}
    return _dol


def convert_dol_to_lod(dict_of_lists):
    """
    converts dict of lists --> list of dicts
    """
    _lod = [dict(zip(dict_of_lists, t)) for t in zip(*dict_of_lists.values())]
    return _lod


def convert_lod_to_md_table(list_of_dicts, table_format="pipe", float_fmt=".3E"):
    ftm_tbl = tabulate(list_of_dicts,
                       headers='keys',
                       tablefmt=table_format,
                       floatfmt=float_fmt)
    return ftm_tbl


def convert_lod_to_pddf(list_of_dicts):
    """
        Converts a list of dictionaries to a pandas dataframe

    Args:
        list_of_dicts: a list of dictionaries

    Returns:
        pandas.DataFrame:

    """
    return pd.DataFrame(list_of_dicts,
                       columns=[*list_of_dicts[0]])


def convert_dict_to_quantity_value(dict):
    """
    Args:
        dict (dict): a dictionary

    Returns:
        list: a list of dicts with quantity-value pairs

    """
    return [{'quantity': k, 'value': dict[k]} for k in dict]



# -*- coding: utf-8 -*-
import os
import numpy
import math
import matplotlib.pyplot as plt
from rmtk.vulnerability.common import utils

def generate_capacity_curves(Sa_mean,Sa_cov,Sd_mean,Sd_cov,type_dist,no_cc,Sa_corr,Sd_corr,SaSd_corr,truncation):
#Creates a set of randomly-generated capacity curves, based on a mean curve.
      
    no_points = len(Sa_mean)
    mean, covariance = calculate_variables(Sa_mean,Sd_mean,Sa_cov,Sd_cov,Sa_corr,Sd_corr,SaSd_corr)
    zero_col = numpy.zeros((no_cc,1))
    Sa = []
    Sd = []
    generated_cc = 0
        
    while generated_cc < no_cc:
        sample = numpy.random.multivariate_normal(mean,covariance,1)
        if validate_sample(sample[0],truncation,Sa_mean,Sa_cov,Sd_mean,Sd_cov):
            Sa.append(sample[0][0:no_points])
            Sd.append(sample[0][no_points:])
            generated_cc = generated_cc +1
    
    Sa = numpy.array(Sa)
    Sd = numpy.array(Sd)
    Sa_corr = numpy.corrcoef((Sa[:,-1],Sa[:,-2]))
    Sd_corr = numpy.corrcoef((Sd[:,-1],Sd[:,-2]))
    Sa_Sd_corr = numpy.corrcoef((Sa[:,-1],Sd[:,-1]))
    
    print 'Actual correlation in Sa: '+str(Sa_corr[0][1])
    print 'Actual correlation in Sd: '+str(Sd_corr[0][1])
    print 'Actual correlation between Sa and Sd: '+str(Sa_Sd_corr[0][1])

    Sa = numpy.hstack((zero_col,Sa))
    Sd = numpy.hstack((zero_col,Sd))
    
    capacity_curves = {'type': None, 'periods': None, 'heights': None, 
        'gamma': None, 'Sdy': None, 'Say': None, 'Sd': None, 'Sa': None}

    capacity_curves['type'] = 'Sd-Sa'
    capacity_curves['Sd'] = Sd
    capacity_curves['Sa'] = Sa    
    
    return capacity_curves
    
def validate_sample(sample,truncation,Sa_mean,Sa_cov,Sd_mean,Sd_cov):
#This function performs a series of sanity checks in order to validate the randomly generated
#sample
    
    result = True
    no_points = len(Sa_mean)
    Sa = sample[0:no_points]
    Sd = sample[no_points:]
    delta = Sd[1:]-Sd[0:-1]
    Sa_sigma = numpy.multiply(Sa_mean,Sa_cov)
    Sd_sigma = numpy.multiply(Sd_mean,Sd_cov)
    Sa_epsilon = abs(numpy.divide((Sa-Sa_mean),Sa_sigma))
    Sd_epsilon = abs(numpy.divide((Sd-Sd_mean),Sd_sigma))

    if min(sample) < 0:
        result = False

    if min(delta) <= 0:
        result = False
        
    if max(Sa_epsilon) > truncation:
        result = False
        
    if max(Sd_epsilon) > truncation:
        result = False
    
    return result
    
def calculate_variables(Sa_mean,Sd_mean,Sa_cov,Sd_cov,Sa_corr,Sd_corr,SaSd_corr):
#This function returns the mean values of Sa and Sd values and calculates the covariance matrix  
   
    mean = []
    for value in Sa_mean:
        mean.append(value)
    for value in Sd_mean:
        mean.append(value)  

    cov = []
    for value in Sa_cov:
        cov.append(value)
    for value in Sd_cov:
        cov.append(value)
        
    no_points = len(mean)
    covariance = numpy.zeros((no_points,no_points))        
    
    for irow in range(no_points):
        for icol in range(no_points):
            sigma1 = mean[irow]*cov[irow] 
            sigma2 = mean[icol]*cov[icol]
            if irow == icol:
                covariance[irow][icol] = sigma1*sigma2
            else:
                if irow < no_points/2 and icol < no_points/2:
                    corr =  Sa_corr
                elif irow < no_points/2 and icol >= no_points/2:
                    corr =  SaSd_corr
                elif irow >= no_points/2 and icol < no_points/2:
                    corr =  SaSd_corr
                elif irow >= no_points/2 and icol >= no_points/2:
                    corr =  Sd_corr
                
                covariance[irow][icol] = sigma1*sigma2*corr
    return mean, covariance
    
def add_information(capacity_curves,attribute,type,data):
#Adds additional information to the capacity_curves dictionary (like storey heights, period,participation
#factors,etc.)
    
    no_capacity_curves = len(capacity_curves['Sa'])
    
    if attribute == 'heights' or attribute == 'periods' or attribute == 'gamma':           
        if type == 'value':
            values = []
            for icc in range(no_capacity_curves):
                values.append(data)
            capacity_curves[attribute] = values
        elif type == 'vector':
            capacity_curves[attribute] = data
        elif type == 'calculate' and attribute == 'periods': 
            periods = []
            for icc in range(no_capacity_curves):
                Sd = capacity_curves['Sd'][icc][data] 
                Sa = capacity_curves['Sa'][icc][data] 
                periods.append(2*math.pi*math.sqrt(Sd/(Sa*9.81)))
            capacity_curves[attribute] = periods
            
    elif attribute == 'yielding point':
        Sdy = []
        Say = []
        for icc in range(no_capacity_curves):
            Sdy.append(capacity_curves['Sd'][icc][data])
            Say.append(capacity_curves['Sa'][icc][data])
        capacity_curves['Sdy'] = Sdy 
        capacity_curves['Say'] = Say 
    else:
        print attribute + ' is not a recognized attribute. No information was added.'
        
    return capacity_curves


import pandas as pd
from . import RectangularBeamProperties as rect_props


def os2d_general():
    _str = '# ---------------------------------------------------------------------------------------------------\n'
    _str += '# U N I T S\n'
    _str += '# ---------------------------------------------------------------------------------------------------\n'
    _str += '# Length : m\n'
    _str += '# Force  : kN\n'
    _str += '# Moment : kNm\n'
    _str += '# Stress : kPa\n'
    _str += '# Mass   : ton\n'
    _str += '\n'
    _str += '# Model domain 3DOF\n'
    _str += '\n'
    _str += 'model BasicBuilder -ndm 2 -ndf 3\n'
    return _str + '\n'


def os2d_nodes(_df_nodes):
    _str = '# ---------------------------------------------------------------------------------------------------\n'
    _str += '# N O D E S\n'
    _str += '# ---------------------------------------------------------------------------------------------------\n'
    _str += '#        nodeID            X            Y\n'
    for row in _df_nodes.itertuples():
        _str += f'node {row.Node:10} {row.X:12.3f} {row.Z:12.3f}\n'
    return _str + '\n'


# Στηρίξεις
def os2d_node_restraints(_df_node_restraints):
    _str = '# ---------------------------------------------------------------------------------------------------\n'
    _str += '# R E S T R A I N T S\n'
    _str += '# ---------------------------------------------------------------------------------------------------\n'
    _str += '# fix    nodeID   dxFixity   dyFixity    rzFixity\n'
    for row in _df_node_restraints.itertuples():
        _str += f'fix {row.Node:10} {row.U1:10} {row.U3:10}  {row.R2:10} \n'
    return _str + '\n'


# Διαφράγματα
def os2d_diaphragms(_df_node_constraints):
    _str = '# ---------------------------------------------------------------------------------------------------\n'
    _str += '# C O N S T R A I N T S\n'
    _str += '# ---------------------------------------------------------------------------------------------------\n'
    _str += '#          master     slave      dof\n'
    df_diaphragms = _df_node_constraints[_df_node_constraints.TYPE == 'Diaphragm']
    # print(df_diaphragms)
    for row in df_diaphragms.itertuples():
        nodes = row.NODES.split(',')
        #     print(f'masternode for diaph {row.ID} is {nodes[0]}')
        for node in nodes:
            if node != nodes[0]:
                _str += f'equalDOF {nodes[0]:>8}  {node:>8}        1\n'
    return _str + '\n'


# Μάζες στους κόμβους
def os2d_masses_nodal(_df_masses_nodal):
    _str = '# ---------------------------------------------------------------------------------------------------\n'
    _str += '# M A S S E S\n'
    _str += '# ---------------------------------------------------------------------------------------------------\n'
    _str += '#Mass Definition : mass $NodeTag $(ndf nodal mass values corresponding to each DOF)\n'
    for row in _df_masses_nodal.itertuples():
        _str += f'mass {row.NODE:>10} {row.UX:>10.3f} {row.UY:>10.3f}{row.RZ:>10.3f}\n'
    return _str + '\n'


# Connectivity - Frame
def os2d_frame_connectivity(_df_frame_connectivity, _df_section_rect_elastic, _df_materials_general):
    _str = '# ---------------------------------------------------------------------------------------------------\n'
    _str += '# E L A S T I C   B E A M - C O L U M N   E L E M E N T S\n'
    _str += '# ---------------------------------------------------------------------------------------------------\n'
    _str += '# Elastic Beam Column Definition\n'
    _str += '# element elasticBeamColumn $eleTag $iNode $jNode $A $E $Iz $transfTag <-mass $MassPerUnitLength>\n'
    for row in _df_frame_connectivity.itertuples():
        _sect_geom = row.SECT_GEOM
        _sect_id = row.SECTION_ID
        _sect_props = get_section_properties(_sect_geom, _sect_id, _df_section_rect_elastic, _df_materials_general)
        _str += f'element elasticBeamColumn{row.FRAME:>7} {row.NODEI:>7} {row.NODEJ:>7}' \
                f'{_sect_props[2]:>12.6f} {_sect_props[0]:>14.1f} {_sect_props[3]:>15.10f}' \
                f'    1   -mass    0.000\n'
    return _str + '\n'


def get_section_properties(sect_geom, sect_id, _df_section_rect_elastic, _df_materials_general):
    if sect_geom == 'RECT':
        _df = _df_section_rect_elastic.set_index("RECT_SECTION_ID")
        _row = _df.loc[sect_id]
        _mat = _row.MATERIAL
        _b = _row.b
        _h = _row.h
        _mod_area = _row.modArea
        _mod_shear = _row.modShear
        _mod_moment = _row.modMoment
        _mod_torsional = _row.modTorsional
        _rect_props = rect_props.RectangularBeamProperties(_b, _h)

        _df_mat = _df_materials_general.set_index("MATERIAL_ID")
        _mat_E = _df_mat.at[_mat, 'E']
        _mat_poisson = _df_mat.at[_mat, 'POISSON']
        _mat_G = _mat_E / (2 * (1.0 + _mat_poisson))

        _tup = _mat_E, \
               _mat_G, \
               _rect_props.area() * _mod_area, \
               _rect_props.moment_of_inertia_xx() * _mod_moment, \
               _rect_props.shear_area_2() * _mod_shear
        # returns a tuple with: E, G, A, Iz, Avy
        # A, Iz, Avy are multiplied with mod factors
        return _tup



"""

    .. uml::

        class ratios <<(M,#FF7700)>> {
        .. functions ..
        ρ()
        ρw()
        ω()
        }

"""

import math
from . import areas


def ρ(As, b, d):
    """
        Ποσοστό διαμήκους οπλισμού

    Args:
        As (float): Reinforcement area
        b (float): Section width
        d (float): Section effective depth

    Returns:
        float:
    """
    return As / (b * d)


def ρw(nw, diaw, b, s, α=math.pi/2):
    """
        Ποσοστό εγκάρσιου οπλισμού

    Args:
        nw (float): Τμήσεις συνδετήρα
        diaw (float): Διάμετρος συνδετήρα
        b (float): Section width
        s (float): Space between transverse reinforcement
        α (float): Transverse reinforcement angle

    Returns:
        float:
    """
    Asw = nw * areas.As(diaw)
    return Asw / (b * s * math.sin(α))


def ρw_vol(hoops_length, diaw, b, h, s):
    """
        Ογκομετρικό ποσοστό εγκάρσιου οπλισμού

    Args:
        hoops_length (float): Συνολικό μήκος σκελών συνδετήρων
        diaw (float): Διάμετρος συνδετήρα
        b (float): Section width
        h (float): Section height
        s (float): Space between transverse reinforcement

    Returns:
        float:
    """
    return hoops_length * areas.As(diaw) / (b * h * s)



def ω(As: float, b: float, d: float, fy: float, fc: float):
    """
        Μηχανικά ποσοστά οπλισμού

    Args:
        As (float): Reinforcement area
        b (float): Section width
        d (float): Section effective depth
        fy (float):
        fc (float):

    Returns:
        float:
    """
    return As / (b * d) * (fy / fc)



import numpy as np
import math
from ..common.io.output import OutputTable, OutputString, OutputExtended
from dataclasses import dataclass
from ..common.math.charts.linexy import LineXY
from ..common.math.numerical import xy_with_endpoints

@dataclass
class BilinearCurve:
    """
        .. uml::

            class BilinearCurve {
            .. members ..
            + ay: float
            + au: float
            + dy: float
            + du: float
            + a0: float = 0.
            + d0: float = 0.
            .. properties ..
            + μ()
            + kel()
            + kinel()
            + α()
            + β0()
            + d_array()
            + a_array()
            + T0
            + all_quantities
            .. methods ..
            + Teq(Τ0)

            }
    """
    ay: float
    au: float
    dy: float
    du: float
    a0: float = 0.
    d0: float = 0.

    @property
    def μ(self) -> float:
        """
        .. math::
            μ = \dfrac{d_u - d_0}{d_y - d_0}

        """
        return (self.du-self.d0) / (self.dy-self.d0)

    @property
    def kel(self) -> float:
        """
        Slope of the elastic branch, :math:`k_{el} = \dfrac{a_y - a_0}{d_y - d_0}`
        """
        return (self.ay - self.a0) / (self.dy - self.d0)

    @property
    def kinel(self) -> float:
        """
        Slope of the inelastic branch, :math:`k_{inel} = \dfrac{a_u - a_y}{d_u - d_y}`
        """
        return (self.au - self.ay) / (self.du - self.dy)

    @property
    def α(self) -> float:
        """
        Hardening, :math:`α = k_{inel}/k_{el}`
        """
        return self.kinel / self.kel

    @property
    def β0(self) -> float:
        """
        Damping after Fema440 Equation 3-6

        .. math::
            β_0 = \dfrac{2}{π}\dfrac{(μ - 1) \cdot (1 - α)}{μ \cdot (1 + α \cdot μ - α)}
        """
        μ = self.μ
        α = self.α
        return (2 / math.pi) * ((μ - 1) * (1 - α)) / (μ * (1 + α * μ - α))

    @property
    def d_array(self):
        return np.array([self.d0, self.dy, self.du])

    @property
    def a_array(self):
        return np.array([self.a0, self.ay, self.au])

    @property
    def T0(self):
        """
        Initial period T0

        .. math::
            T_0 = 2 π \sqrt{\dfrac{d_y - d_0}{a_y - a_0}}
        """
        return 2*math.pi*((self.dy-self.d0)/(self.ay - self.a0))**0.5

    def Teq(self, T0):
        """
        Effective period Teq  after Fema440 Equation 3-5

        .. math::
            T_{eq} = T_0 \sqrt{\dfrac{μ}{1 + α \cdot μ - α}}
        """
        μ = self.μ
        α = self.α
        T0 = self.T0
        if μ <= 1.0:
            return T0
        else:
            return T0 * (μ / (1 + α * μ - α)) ** 0.5

    @property
    def all_quantities(self):
        out = OutputTable()
        out.data.append({'quantity': 'x_0', 'value': self.d0})
        out.data.append({'quantity': 'x_y', 'value': self.dy})
        out.data.append({'quantity': 'x_u', 'value': self.du})
        out.data.append({'quantity': 'y_0', 'value': self.a0})
        out.data.append({'quantity': 'y_y', 'value': self.ay})
        out.data.append({'quantity': 'y_u', 'value': self.au})
        out.data.append({'quantity': 'kel', 'value': self.kel})
        out.data.append({'quantity': 'kinel', 'value': self.kinel})
        out.data.append({'quantity': 'μ', 'value': self.μ})
        out.data.append({'quantity': 'α', 'value': self.α})
        return out

@dataclass
class Bilin:
    """
        .. uml::

            class Bilin {
            .. members ..
            + curve_ini: LineXY = LineXY()
            + xtarget: float = 0.0
            + dropstrength: float = 0.75
            + elastoplastic: bool = False
            + allowa010: bool = True
            + EPSILON: float = 0.000001
            .. properties ..
            + x_max()
            + x_min()
            + y_max()
            + y_min()
            + y_max_index()
            .. methods ..
            + calc()
            - __iteration()
            - __curve_to_xcheck()
            - __get_area()
            }
    """
    curve_ini: LineXY = LineXY()
    xtarget: float = 0.0
    dropstrength: float = 0.75
    elastoplastic: bool = False
    allowa010: bool = True
    # x_ini: np.array = np.array([])
    # y_ini: np.array = np.array([])
    EPSILON: float = 0.000001

    def __post_init__(self):
        self.bilinear_curve = BilinearCurve(0., 0., 0., 0.)
        self.output = OutputExtended()
        self.output.outputStrings['main'] = OutputString(data=['Bilin main output'])
        self.output.outputTables['InitialCurve'] = OutputTable()
        self.output.outputTables['Iterations'] = OutputTable()
        self.output.outputTables['BilinearCurve'] = OutputTable()

    # def load_space_delimited(self, fname, delimiter=' '):
    #     self.x_ini, self.y_ini = np.loadtxt(fname, delimiter=delimiter, usecols=(0, 1), unpack=True)

    # def load_space_delimited_string(self, text, delimiter=' '):
    #     self.x_ini = np.array([], dtype=np.float)
    #     self.y_ini = np.array([], dtype=np.float)
    #     spl = text.splitlines()
    #     for ln in spl:
    #         xy = ln.split(delimiter)
    #         self.x_ini = np.append(self.x_ini, xy[0])
    #         self.y_ini = np.append(self.y_ini, xy[1])
    #     self.x_ini = self.x_ini.astype(float)
    #     self.y_ini = self.y_ini.astype(float)

    @staticmethod
    def __curve_to_xcheck(x, y, xtarget):
        if xtarget == 0.0:
            return x, y
        return xy_with_endpoints(x, y, x[0], xtarget)

    @staticmethod
    def __get_area(x, y):
        return np.trapz(y, x)

    def calc(self):

        for i in range(len(self.curve_ini.x)):
            self.output.outputTables['InitialCurve'].data.append({'x': self.curve_ini.x[i],
                                                                  'y': self.curve_ini.y[i]})

        # Βρίσκω την αρχική μετατόπιση σε περίπτωση που δεν είναι 0
        x_ini0 = self.curve_ini.x[0]

        self.output.outputStrings['main'].data.append('')
        self.output.outputStrings['main'].data.append(f'Αρχική μετατόπιση: {x_ini0}')

        ymax = self.curve_ini.y_max
        i_ymax = self.curve_ini.y_max_index

        y_max_to_end = self.curve_ini.y[i_ymax:]
        x_max_to_end = self.curve_ini.x[i_ymax:]

        i = 0
        while y_max_to_end[i] > self.dropstrength * ymax and i < len(y_max_to_end) - 1:
            i += 1
        x_for_dropstrength = x_max_to_end[i]


        x_target = min(x_for_dropstrength, self.xtarget)

        # Κρατώ την καμπύλη μέχρι το xtarget, αν υπάρχει
        x_xcheck, y_xcheck = self.__curve_to_xcheck(self.curve_ini.x, self.curve_ini.y, x_target)

        # Αφαιρώ την αρχική μετατόπιση ώστε η καμπύλη να ξεκινά από το (0, 0)
        x_xcheck = x_xcheck - x_ini0

        # self.output.log_text.append('')
        # self.output.log_text.append('X values μέχρι το xtarget, αφαιρώντας (αν υπάρχει) την αρχική μετατόπιση')
        # self.output.log_text.append(str(x_xcheck))
        # self.output.log_text.append('')
        # self.output.log_text.append('Y values μέχρι το xtarget')
        # self.output.log_text.append(str(y_xcheck))

        # Βρίσκω τις δυσκαμψίες σε κάθε βήμα
        # Αλλάχω προσωρινά το x(0) για να μη διαιρεί με 0
        # x_xcheck[0] = self.EPSILON
        # k = np.divide(y_xcheck, x_xcheck)
        # x_xcheck[0] = 0.0

        # self.output.log_text.append('')
        # self.output.log_text.append('Δυσκαμψίες (y(i)/x(i)')
        # self.output.log_text.append(str(k))

        y02 = 0.2 * max(y_xcheck)
        x02 = float(np.interp(y02, y_xcheck, x_xcheck))

        k02 = y02 / x02

        self.output.outputStrings['main'].data.append('')
        self.output.outputStrings['main'].data.append('Έλεγχος στο 20% του ymax')
        self.output.outputStrings['main'].data.append(f'x(02)={x02}, y(02)={y02}. Οπότε k(02)={k02}')

        # Βρίσκω το εμβαδό
        area = self.__get_area(x_xcheck, y_xcheck)
        self.output.outputStrings['main'].data.append('')
        self.output.outputStrings['main'].data.append(f'Εμβαδό καμπύλης: {area}')

        # _iterations_dict = NamedLogTable('Iterations')
        iteration_number = 0
        kel = k02
        error = 100.
        while error > self.EPSILON:
            iteration_number += 1  # This is the same as count = count + 1

            x_y, y_y, x_u, y_u, kinel, k_06 = self.__iteration(x_xcheck, y_xcheck, kel, area)
            error = np.abs((kel - k_06) / k_06)

            self.output.outputTables['Iterations'].data.append({'__iteration': iteration_number,
                                                                'x_y': x_y,
                                                                'y_y': y_y,
                                                                'x_u': x_u,
                                                                'y_u': y_u,
                                                                'kinel': kinel,
                                                                'kel': kel,
                                                                'k_06': k_06,
                                                                'error': error})


            self.bilinear_curve = BilinearCurve(ay=y_y,
                                                au=y_u,
                                                dy=x_y+x_ini0,
                                                du=x_u+x_ini0,
                                                a0=0.,
                                                d0=+x_ini0)
            kel = k_06

            if iteration_number > 1000:
                break

        for i in range(len(self.bilinear_curve.d_array)):
            self.output.outputTables['BilinearCurve'].data.append({'x': self.bilinear_curve.d_array[i],
                                                                   'y': self.bilinear_curve.a_array[i]})

    def __iteration(self, x, y, kel, area):
        rcount = len(x) - 1
        ymax = max(y)

        # ********** Αρχικός υπολογισμός *****************
        # if y[rcount - 1] >= (2 * ymax + y[rcount]) / 3.0 and y[rcount] <= (2 * ymax + y[rcount]) / 3.0:
        #     y_u = y[rcount - 1]
        # else:
        #     y_u = (2 * ymax + y[rcount]) / 3.0
        y_u = (2 * ymax + y[rcount]) / 3.0

        x_u = x[rcount]

        x_y = (2 * area - x[rcount] * y_u) / (kel * x[rcount] - y_u)
        y_y = kel * x_y

        kinel = (y_u - y_y) / (x_u - x_y)

        # ********** 2η περίπτωση kinel/kel>0.1 *****************
        if self.allowa010 == False and kinel / kel > 0.1:
            alpha = 1.
            beta = -2. * x_u * kel
            gamma = 1.8 * area * kel + 0.1 * (kel * x_u) ** 2
            y_u = (-beta - (beta * beta - 4 * alpha * gamma) ** 0.5) / (2. * alpha)
            x_y = (y_u - 0.1 * kel * x_u) / (0.9 * kel)
            y_y = kel * x_y

            # ********** 3η περίπτωση Ελαστοπλαστικό ή kinel<0 *****************
        if self.elastoplastic == True or kinel < 0.0:
            alpha = 1.
            beta = -2. * x_u
            gamma = 2. * area / kel
            x_y = (-beta - (beta * beta - 4 * alpha * gamma) ** 0.5) / (2. * alpha)
            y_y = x_y * kel
            y_u = y_y

        y_06 = 0.6 * y_y
        x_06 = float(np.interp(y_06, y, x))
        k_06 = y_06 / x_06

        return x_y, y_y, x_u, y_u, kinel, k_06


    # def results_kel(self):
    #     return self.y_results[1] / self.x_results[1]
    #
    # def results_kinel(self):
    #     return (self.y_results[2] - self.y_results[1]) / (self.x_results[2] - self.x_results[1])
    #
    # def results_ductility(self):
    #     return self.x_results[2] / self.x_results[1]
    #
    # def results_hardening(self):
    #     return self.results_kinel()/self.results_kel()


    def __str__(self):
        # return ('\n').join((self.output.log_text.text_list))
        return self.output.outputStrings['main'].__str__()



from dataclasses import dataclass, field
from typing import List
import numpy as np
import pandas as pd
import h5py
from scipy.stats import lognorm

import matplotlib.pyplot as plt
import seaborn as sns
sns.set_style("whitegrid")


def p_ds(x, medians, stddevs, dist_type='lognormal'):
    if dist_type == 'lognormal':
        return lognorm(s=stddevs, scale=medians).cdf(x)
    else:
        return None


@dataclass
class FragilityCurves:
    θs: List[float] = field(default_factory=list)
    βs: List[float] = field(default_factory=list)
    μXs: List[float] = field(default_factory=list)
    σXs: List[float] = field(default_factory=list)
    ds_descriptions: List[str] = field(default_factory=list)
    thresholds: List[float] = field(default_factory=list)
    centrals: List[float] = field(default_factory=list)
    __centrals: List[float] = field(default_factory=list, init=False, repr=False)

    __n_ds: int = 0
    typology: str = ''
    description: str = ''
    shape: str = 'lognormal'
    imt: str = 'PGA'
    units: str = 'g'
    df_all: pd.DataFrame = None
    # means: list = field(default_factory=list)

    def __post_init__(self):
        pass

    # -------------------------------------
    # ---- Properties ---------------------
    # -------------------------------------
    @property
    def centrals(self) -> List[float]:
        return self.__centrals

    @centrals.setter
    def centrals(self, value: List[float]):
        self.__centrals = value

    @property
    def n_ds(self) -> int:
        return self.__n_ds

    # -------------------------------------
    # ---- Classmethods -------------------
    # -------------------------------------

    @classmethod
    def from_hdf5_θβ_continuous(cls, filename, group_name, dataset_name):
        with h5py.File(filename, mode='r') as f:
            group = f.get(group_name)
            dataset = group[dataset_name]

            fc = cls()
            fc.θs = list(dataset['θ'])
            fc.βs = list(dataset['β'])
            fc.thresholds = list(dataset['thresholds'])
            fc.ds_descriptions = dataset.attrs['damage state description']
            fc.typology = dataset.attrs['typology']
            fc.shape = dataset.attrs['shape']
            fc.imt = dataset.attrs['imt']
            fc.units = dataset.attrs['units']

            # fc.__calc_centrals()

            return fc




    def get_centrals_from_thresholds(self):
        _centrals = []
        _thresholds = self.thresholds.copy()
        _thresholds.append(1.0)
        for i in range(0, len(_thresholds) - 1):
            _centrals.append(0.5 * (_thresholds[i] + _thresholds[i + 1]))

        # Για το DS0 θεωρώ ότι ο κεντρικός δείκτης βλάβης είναι 0
        self.__centrals = [0.0] + _centrals


    # @property
    # def centrals(self):
    #     return self.__centrals
    #
    # @centrals.setter
    # def centrals(self, centrals):
    #     self.__centrals = centrals
    #     # if len(centrals) > 0:
    #     #     self.__centrals = centrals
    #     # else:
    #     #     self.__calc_centrals()

    @property
    def make_the_dataframe_from_θβs(self):
        minIML = 0.0
        maxIML = 4.0
        xs = np.linspace(minIML, maxIML, 500)

        for iDS in range(len(self.θs)):
            _θ = self.θs[iDS]
            _β = self.βs[iDS]
            p_ds(xs, _θ, _β, self.dist_type)


    def get_μs_from_medians(self):
        if self.dist_type == 'lognormal':
            self.μXs = np.log(np.array(self.θs)).tolist()
            
    # def get_means_from_μs(self):
    #     if self.dist_type == 'lognormal':
    #         self.means = np.exp(np.array(self.μs) + 0.5*np.array(self.stddevs)**2).tolist()



    def calc(self, x):
        p = p_ds(x, self.medians, self.stddevs, self.dist_type)

        _δP = []
        for i in range(0, len(p) - 1):
            _δP.append(p[i] - p[i + 1])
        # Για το τελευταίο DS η πιθανότητα είναι απευθείας από την καμπύλη τρωτότητας
        _δP.append(p[-1])
        # Για το DS0 η πιθανότητα είναι η 1.0 μείον την πιθανότητα του DS1
        δP = [1.0 - p[0]] + _δP

        if len(self.thresholds) > 0:
            mdf = sum(np.array(δP) * np.array(self.centrals))
        else:
            mdf = None

        _results = {'P': p,
                    'δP': δP,
                    'MDF': mdf}

        return _results

    def plot_fragility_model(self, minIML, maxIML, dist_type='lognormal', colors=None):
        xs = np.linspace(minIML, maxIML, 500)

        f, ax = plt.subplots(figsize=(12, 8))

        for iDS in range(len(self.θs)):
            _θ = self.θs[iDS]
            _β = self.βs[iDS]

            if colors is None:
                color = 'grey'
            else:
                if colors=='default5':
                    colors = ['#abdb57', '#086c34', '#e28b05', '#fb4c4c', '#b91f1f']
                color = colors[iDS]

            if len(self.ds_descriptions) > 0:
                _label = f'{self.ds_descriptions[iDS]}'
            else:
                _label = f'DS{iDS + 1}'

            ax.plot(xs, p_ds(xs, _θ, _β, dist_type), linewidth=2, color=color, label=_label)

        ax.set_xlim(minIML, maxIML)

        ax.set_xlabel(f'{self.imt} [{self.units}]', fontsize=12)

        ax.set_ylim(0., 1.01)
        ax.set_ylabel('Probabilty of exceedance', fontsize=12)
        ax.legend()

        return f, ax










from dataclasses import dataclass, field
from typing import List  # , Dict, Tuple
from streng.phd.building_models.model2d_older.model_input import Model2d

@dataclass
class EtoolsOutput:
    input_log_filename: str
    log_lines: List[str] = field(default_factory=list)
    title_lines: List[tuple] = field(default_factory=list)
    column_bending_reinforcement: List[dict] = field(default_factory=list)
    column_shear_reinforcement: List[dict] = field(default_factory=list)
    beam_bending_reinforcement: List[dict] = field(default_factory=list)
    beam_shear_reinforcement: List[dict] = field(default_factory=list)

    model: Model2d = None


    def __post_init__(self):
        f = open(self.input_log_filename, 'r')
        self.log_lines = f.readlines()
        self.title_lines = [(i, s) for i, s in enumerate(self.log_lines) if '--- ' in s and ' ---' in s]
        self.get_column_bending_reinforcement()
        self.get_beam_bending_reinforcement()
        self.get_beam_shear_reinforcement()
        self.get_column_shear_reinforcement()

    def get_column_bending_reinforcement(self):
        # Βρίσκω αρχικά τον αύξοντα αριθμό του τίτλου και τη γραμμή του
        _title_index_col_bend, _line_col_bend = \
            [(i, x[0]) for i, x in enumerate(self.title_lines) if "Οπλισμοί κάμψης υποστυλωμάτων" in x[1]][0]
        _line_next_title = self.title_lines[_title_index_col_bend+1][0]

        _col_bending_block = self.log_lines[_line_col_bend:_line_next_title]
        _col_min_reinf_lines = [i for i, s in enumerate(_col_bending_block) if 'Ελάχιστοι οπλισμοι' in s]
        # print(f'_title_index_col_bend = {_title_index_col_bend}')
        # print(f'_line_col_bend = {_line_col_bend}')
        # print(f'_line_next_title = {_line_next_title}')
        _column_bending_reinforcement = []

        for i in _col_min_reinf_lines:
            _col_id_line = _col_bending_block[i-2].split()
            _col_id = _col_id_line[0] + '-' + _col_id_line[1]

            _Asminbot = _col_bending_block[i].split()[2]
            _Asmaxbot = _col_bending_block[i+1].split()[2]
            _Ascalcbot = _col_bending_block[i+2].split()[2]
            _Asreqbot = _col_bending_block[i + 3].split()[2]
            _Assuggestedbot = _col_bending_block[i + 4].split()[2]

            _Asmintop = _col_bending_block[i].split()[3]
            _Asmaxtop = _col_bending_block[i+1].split()[3]
            _Ascalctop = _col_bending_block[i+2].split()[3]
            _Asreqtop = _col_bending_block[i + 3].split()[3]
            _Assuggestedtop = _col_bending_block[i + 4].split()[3]

            _column_bending_reinforcement.append({'col_id': _col_id,
                                                  'Asminbot': _Asminbot,
                                                  'Asmaxbot': _Asmaxbot,
                                                  'Ascalcbot': _Ascalcbot,
                                                  'Asreqbot': _Asreqbot,
                                                  'Assuggestedbot': _Assuggestedbot,
                                                  'Asmintop': _Asmintop,
                                                  'Asmaxtop': _Asmaxtop,
                                                  'Ascalctop': _Ascalctop,
                                                  'Asreqtop': _Asreqtop,
                                                  'Assuggestedtop': _Assuggestedtop,
                                                  })

        self.column_bending_reinforcement = _column_bending_reinforcement

    def get_beam_bending_reinforcement(self):
        # Βρίσκω αρχικά τον αύξοντα αριθμό του τίτλου και τη γραμμή του
        _title_index_beam_bend, _line_beam_bend = \
            [(i, x[0]) for i, x in enumerate(self.title_lines) if "Οπλισμοί κάμψης δοκών" in x[1]][0]
        _line_next_title = self.title_lines[_title_index_beam_bend + 1][0]

        _beam_bending_block = self.log_lines[_line_beam_bend:_line_next_title]
        _beam_min_reinf_lines = [i for i, s in enumerate(_beam_bending_block) if 'Ελάχιστοι οπλισμοι' in s]

        _beam_bending_reinforcement = []

        for i in _beam_min_reinf_lines:
            _beam_id_line = _beam_bending_block[i - 2].split()
            _beam_id = _beam_id_line[0] + '-' + _beam_id_line[1]

            _Asmin_leftbot = _beam_bending_block[i].split()[2]
            _Asmax_leftbot = _beam_bending_block[i + 1].split()[2]
            _Ascalc_leftbot = _beam_bending_block[i + 2].split()[2]
            _Asreq_leftbot = _beam_bending_block[i + 3].split()[2]
            _Assuggested_leftbot = _beam_bending_block[i + 4].split()[2]

            _Asmin_lefttop = _beam_bending_block[i].split()[3]
            _Asmax_lefttop = _beam_bending_block[i + 1].split()[3]
            _Ascalc_lefttop = _beam_bending_block[i + 2].split()[3]
            _Asreq_lefttop = _beam_bending_block[i + 3].split()[3]
            _Assuggested_lefttop = _beam_bending_block[i + 4].split()[3]
            
            _Asmin_rightbot = _beam_bending_block[i].split()[6]
            _Asmax_rightbot = _beam_bending_block[i + 1].split()[6]
            _Ascalc_rightbot = _beam_bending_block[i + 2].split()[6]
            _Asreq_rightbot = _beam_bending_block[i + 3].split()[6]
            _Assuggested_rightbot = _beam_bending_block[i + 4].split()[6]

            _Asmin_righttop = _beam_bending_block[i].split()[7]
            _Asmax_righttop = _beam_bending_block[i + 1].split()[7]
            _Ascalc_righttop = _beam_bending_block[i + 2].split()[7]
            _Asreq_righttop = _beam_bending_block[i + 3].split()[7]
            _Assuggested_righttop = _beam_bending_block[i + 4].split()[7]

            _beam_bending_reinforcement.append({'beam_id': _beam_id,
                                                'Asmin_leftbot': _Asmin_leftbot,
                                                'Asmax_leftbot': _Asmax_leftbot,
                                                'Ascalc_leftbot': _Ascalc_leftbot,
                                                'Asreq_leftbot': _Asreq_leftbot,
                                                'Assuggested_leftbot': _Assuggested_leftbot,
                                                'Asmin_lefttop': _Asmin_lefttop,
                                                'Asmax_lefttop': _Asmax_lefttop,
                                                'Ascalc_lefttop': _Ascalc_lefttop,
                                                'Asreq_lefttop': _Asreq_lefttop,
                                                'Assuggested_lefttop': _Assuggested_lefttop,

                                                'Asmin_rightbot': _Asmin_rightbot,
                                                'Asmax_rightbot': _Asmax_rightbot,
                                                'Ascalc_rightbot': _Ascalc_rightbot,
                                                'Asreq_rightbot': _Asreq_rightbot,
                                                'Assuggested_rightbot': _Assuggested_rightbot,
                                                'Asmin_righttop': _Asmin_righttop,
                                                'Asmax_righttop': _Asmax_righttop,
                                                'Ascalc_righttop': _Ascalc_righttop,
                                                'Asreq_righttop': _Asreq_righttop,
                                                'Assuggested_righttop': _Assuggested_righttop,
                                                })

        self.beam_bending_reinforcement = _beam_bending_reinforcement

    def get_beam_shear_reinforcement(self):
        # Βρίσκω αρχικά τον αύξοντα αριθμό του τίτλου και τη γραμμή του
        _title_index_beam_shear, _line_beam_shear = \
            [(i, x[0]) for i, x in enumerate(self.title_lines) if "Οπλισμοί διάτμησης δοκών" in x[1]][0]
        _line_next_title = self.title_lines[_title_index_beam_shear + 1][0]
        _beam_shear_block = self.log_lines[_line_beam_shear:_line_next_title]

        _beam_shear_reinforcement = []
        i = 5  # start line after title
        while len(_beam_shear_block[i].split()) > 2:
            _line = _beam_shear_block[i].split()
            _beam_id = _line[0] + '-' + _line[1]
            _shear_reinf_left = _line[7]
            _shear_reinf_right = _line[13]
            _beam_shear_reinforcement.append({'beam_id': _beam_id,
                                              'shear_reinf_left': _shear_reinf_left,
                                              'shear_reinf_right': _shear_reinf_right,
                                              })
            i += 1
        self.beam_shear_reinforcement = _beam_shear_reinforcement

    def get_column_shear_reinforcement(self):
        # Βρίσκω αρχικά τον αύξοντα αριθμό του τίτλου και τη γραμμή του
        _title_index_col_shear, _line_col_shear = \
            [(i, x[0]) for i, x in enumerate(self.title_lines) if "Οπλισμοί διάτμησης υποστυλωμάτων" in x[1]][0]
        if _title_index_col_shear != len(self.title_lines) - 1:
            _line_next_title = self.title_lines[_title_index_col_shear + 1][0]
        else:
            _line_next_title = len(self.log_lines)
        _col_shear_block = self.log_lines[_line_col_shear:_line_next_title]

        _column_shear_reinforcement = []
        i = 5  # start line after title
        while len(_col_shear_block[i].split()) > 2:
            _line = _col_shear_block[i].split()
            _beam_id = _line[0] + '-' + _line[1]
            _shear_reinf_left = _line[7]
            _shear_reinf_right = _line[14]
            _column_shear_reinforcement.append({'col_id': _beam_id,
                                              'shear_reinf_left': _shear_reinf_left,
                                              'shear_reinf_right': _shear_reinf_right,
                                              })
            i += 1
        self.column_shear_reinforcement = _column_shear_reinforcement



    def write_group_section_reinforcements(self, filename):
        out = []

        out.append('Οπλισμός δοκών')
        out.append('')
        out.extend(self.get_group_beam_bending_reinforcement())

        out.append('Οπλισμός στύλων')
        out.append('')
        out.extend(self.get_group_column_bending_reinforcement())

        file = open(filename, 'w')
        file.write('\n'.join(out))
        file.close()

    def get_group_beam_bending_reinforcement(self):
        out = []

        beam_section_names = [index for index, row in self.model.input_data['sections'].iterrows() if
                              row['ElementType'] == 'BEAM']
        # print(beam_section_names)

        for beam_name in beam_section_names:
            out.append(f'Δοκοί με διατομή: {beam_name}')
            out.append('')

            elements_with_section = [f'STORY{row["Storey"]}-{row["Line_Etabs"]}' for index, row in
                                     self.model.input_data['elements'].iterrows() if row['GeometryName'] == beam_name]

            out.append(f'Οπλισμός κάμψης δοκών με διατομή: {beam_name}')

            out.append(f'{"Στοιχείο":^15}{" || "}{"Αριστερά πάνω":^30}{" | "}{"Αριστερά κάτω":^30}{" || "}' + \
                       f'{"Δεξιά πάνω":^30}{" | "}{"Δεξιά κάτω":^30}{" || "}')
            out.append(f'{"":^15}{" || "}{"Aslt_min":^10}{"Aslt_calc":^10}{"Aslt_sug":^10}{" | "}' + \
                       f'{"Aslb_min":^10}{"Aslb_calc":^10}{"Aslb_sug":^10}{" || "}' + \
                       f'{"Asrt_min":^10}{"Asrt_calc":^10}{"Asrt_sug":^10}{" | "}' + \
                       f'{"Asrb_min":^10}{"Asrb_calc":^10}{"Asrb_sug":^10}{" || "}')

            for el in elements_with_section:
                bbr = next(x for x in self.beam_bending_reinforcement if x['beam_id'] == el)
                out.append(
                    f'{el:^15}{" || "}{bbr["Asmin_lefttop"]:^10}{bbr["Ascalc_lefttop"]:^10}{bbr["Assuggested_lefttop"]:^10}{" | "}' + \
                    f'{bbr["Asmin_leftbot"]:^10}{bbr["Ascalc_leftbot"]:^10}{bbr["Assuggested_leftbot"]:^10}{" || "}' + \
                    f'{bbr["Asmin_righttop"]:^10}{bbr["Ascalc_righttop"]:^10}{bbr["Assuggested_righttop"]:^10}{" | "}' + \
                    f'{bbr["Asmin_rightbot"]:^10}{bbr["Ascalc_rightbot"]:^10}{bbr["Assuggested_rightbot"]:^10}{" || "}')

            out.append(
                '-----------------------------------------------------------------------------------------------------------------------------------------------------------')
            out.append('')

        return out

    def get_group_column_bending_reinforcement(self):
        out = []

        column_section_names = [index for index, row in self.model.input_data['sections'].iterrows() if
                                row['ElementType'] == 'COLUMN']
        # print(column_section_names)

        for column_name in column_section_names:
            out.append(f'Στύλοι με διατομή: {column_name}')
            out.append('')

            elements_with_section = [f'STORY{row["Storey"]}-{row["Line_Etabs"]}' for index, row in
                                     self.model.input_data['elements'].iterrows() if row['GeometryName'] == column_name]

            out.append(f'Οπλισμός κάμψης δοκών με διατομή: {column_name}')

            out.append(f'{"Στοιχείο":^15}{" || "}{"Πόδας":^30}{" | "}{"Κεφαλή":^30}{" || "}')
            out.append(f'{"":^15}{" || "}{"Asb_min":^10}{"Asb_calc":^10}{"Asb_sug":^10}{" | "}' + \
                       f'{"Ast_min":^10}{"Ast_calc":^10}{"Ast_sug":^10}{" || "}')

            for el in elements_with_section:
                cbr = next(x for x in self.column_bending_reinforcement if x['col_id'] == el)
                out.append(
                    f'{el:^15}{" || "}{cbr["Asminbot"]:^10}{cbr["Ascalcbot"]:^10}{cbr["Assuggestedbot"]:^10}{" | "}' + \
                    f'{cbr["Asmintop"]:^10}{cbr["Ascalctop"]:^10}{cbr["Assuggestedtop"]:^10}{" || "}')

            out.append(
                '-----------------------------------------------------------------------------------------------------------------------------------------------------------')
            out.append('')

        return out



#
#
# lumn_bending_reinforcement.append({'col_id': _col_id,
#                                                   'Asminbot': _Asminbot,
#                                                   'Asmaxbot': _Asmaxbot,
#                                                   'Ascalcbot': _Ascalcbot,
#                                                   'Asreqbot': _Asreqbot,
#                                                   'Assuggestedbot': _Assuggestedbot,
#                                                   'Asmintop': _Asmintop,
#                                                   'Asmaxtop': _Asmaxtop,
#                                                   'Ascalctop': _Ascalctop,
#                                                   'Asreqtop': _Asreqtop,
#                                                   'Assuggestedtop': _Assuggestedtop,


from dataclasses import dataclass, field
from streng.common.io.output import OutputTable

@dataclass
class RectangularSectionGeometry:
    b: float
    h: float

    _area: float = field(init=False, repr=False)

    def __post_init__(self):
        self.calc()

    def calc(self):
        self._area = self.b * self.h

    @property
    def area(self) -> float:
        return self._area

    @property
    def moment_of_inertia_yy(self) -> float:
        return self.b ** 3 * self.h / 12

    @property
    def moment_of_inertia_xx(self) -> float:
        return self.h ** 3 * self.b / 12

    @property
    def torsional_constant(self) -> float:
        return self.b ** 3 * self.h * (1 / 3 - 0.21 * self.b / self.h * (1 - self.b ** 4 / (12 * self.h ** 4)))

    @property
    def shear_area_2(self) -> float:
        return 5 / 6 * self._area

    @property
    def shear_area_3(self) -> float:
        return 5 / 6 * self._area

    @property
    def section_modulus_2(self) -> float:
        return self.b ** 2 * self.h / 6

    @property
    def section_modulus_3(self) -> float:
        return self.h ** 2 * self.b / 6

    @property
    def plastic_modulus_2(self) -> float:
        return self.b ** 2 * self.h / 4

    @property
    def plastic_modulus_3(self) -> float:
        return self.h ** 2 * self.b / 4

    @property
    def radius_of_gyration_2(self) -> float:
        return self.b / 12 ** 0.5

    @property
    def radius_of_gyration_3(self) -> float:
        return self.h / 12 ** 0.5

    @property
    def x_g(self) -> float:
        return self.b / 2

    @property
    def y_g(self) -> float:
        return self.h / 2

    @property
    def all_quantities(self):
        out = OutputTable()
        out.data.append({'quantity': 'b', 'value': self.b})
        out.data.append({'quantity': 'h', 'value': self.h})
        out.data.append({'quantity': 'area', 'value': self.area})
        out.data.append({'quantity': 'Iyy', 'value': self.moment_of_inertia_yy})
        out.data.append({'quantity': 'Ixx', 'value': self.moment_of_inertia_xx})
        out.data.append({'quantity': 'J', 'value': self.torsional_constant})
        out.data.append({'quantity': 'ShearArea2', 'value': self.shear_area_2})
        out.data.append({'quantity': 'ShearArea3', 'value': self.shear_area_3})
        out.data.append({'quantity': 'SectionModulus2', 'value': self.section_modulus_2})
        out.data.append({'quantity': 'SectionModulus3', 'value': self.section_modulus_3})
        out.data.append({'quantity': 'PlasticModulus2', 'value': self.plastic_modulus_2})
        out.data.append({'quantity': 'PlasticModulus3', 'value': self.plastic_modulus_3})
        out.data.append({'quantity': 'RadiusOfGyration2', 'value': self.radius_of_gyration_2})
        out.data.append({'quantity': 'RadiusOfGyration3', 'value': self.radius_of_gyration_3})
        out.data.append({'quantity': 'xG', 'value': self.x_g})
        out.data.append({'quantity': 'yG', 'value': self.y_g})
        return out

